@@ -35,25 +35,16 @@
 #include "lowdown.h"
 #include "extern.h"
 
-/*
- * Queue entry for header names.
- * Keep these so we can make sure that headers have a unique "id" for
- * themselves.
- */
-struct	hentry {
-	char		*str; /* header name (raw) */
-	size_t	 	 count; /* references */
-	TAILQ_ENTRY(hentry) entries;
-};
-
 /*
  * Our internal state object.
  */
 struct 	html {
-	TAILQ_HEAD(, hentry) 	 headers_used;
-	ssize_t			 headers_offs; /* header offset */
-	unsigned int 		 flags; /* "oflags" in lowdown_opts */
-	int			 noescape; /* don't escape text */
+	struct hentryq	 	  headers_used; /* headers we've seen */
+	ssize_t			  headers_offs; /* header offset */
+	unsigned int 		  flags; /* "oflags" in lowdown_opts */
+	int			  noescape; /* don't escape text */
+	struct lowdown_buf	**foots; /* footnotes */
+	size_t			  footsz; /* footnotes size  */
 };
 
 /*
@@ -311,66 +302,19 @@ rndr_linebreak(struct lowdown_buf *ob)
 	return HBUF_PUTSL(ob, "<br/>\n");
 }
 
-/*
- * Given the header with non-empty content "header", fill "ob" with the
- * identifier used for the header.
- * This will reference-count the header so we don't have duplicates.
- * Return zero on failure (memory), non-zero on success.
- */
 static int
-rndr_header_id(struct lowdown_buf *ob,
-	const struct lowdown_buf *header, struct html *st)
+rndr_header(struct lowdown_buf *ob, const struct lowdown_buf *content,
+	const struct lowdown_node *n, struct html *st)
 {
-	struct hentry	*hentry;
-
-	/* 
-	 * See if the header was previously already defind. 
-	 * Note that in HTML5, the identifier is case sensitive.
-	 */
-
-	TAILQ_FOREACH(hentry, &st->headers_used, entries) {
-		if (strlen(hentry->str) != header->size)
-			continue;
-		if (strncmp(hentry->str, 
-		    header->data, header->size) == 0)
-			break;
-	}
-
-	/* Convert to escaped values. */
-
-	if (!escape_href(ob, header, st))
-		return 0;
+	ssize_t				 level;
+	const struct lowdown_buf	*buf;
 
 	/*
-	 * If we're non-unique, then append a "count" value.
-	 * XXX: if we have a header named "foo-2", then two headers
-	 * named "foo", we'll inadvertently have a collision.
-	 * This is a bit much to keep track of, though...
+	 * The <hN> level take into account shifteheadinglevelby
+	 * metadata, so offset it here.  Bound us below <h6>.
 	 */
 
-	if (hentry != NULL)
-		return hbuf_printf(ob, "-%zu", ++hentry->count);
-
-	/* Create new header entry. */
-
-	if ((hentry = calloc(1, sizeof(struct hentry))) == NULL)
-		return 0;
-
-	TAILQ_INSERT_TAIL(&st->headers_used, hentry, entries);
-	hentry->count = 1;
-	hentry->str = strndup(header->data, header->size);
-	return hentry->str != NULL;
-}
-
-static int
-rndr_header(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct rndr_header *param, 
-	struct html *st)
-{
-	ssize_t	level;
-
-	level = (ssize_t)param->level + st->headers_offs;
+	level = (ssize_t)n->rndr_header.level + st->headers_offs;
 	if (level < 1)
 		level = 1;
 	else if (level > 6)
@@ -378,19 +322,46 @@ rndr_header(struct lowdown_buf *ob,
 
 	if (ob->size && !hbuf_putc(ob, '\n'))
 		return 0;
+	if (!hbuf_printf(ob, "<h%zu", level))
+		return 0;
 
-	if (content->size && (st->flags & LOWDOWN_HTML_HEAD_IDS)) {
-		if (!hbuf_printf(ob, "<h%zu id=\"", level))
+	/*
+	 * Identifiers can either come from header attributes or as
+	 * computed from the content of the header.
+	 */
+
+	if (n->rndr_header.attr_id.size) {
+		if (!HBUF_PUTSL(ob, " id=\""))
 			return 0;
-		if (!rndr_header_id(ob, content, st))
+		if (!escape_href(ob, &n->rndr_header.attr_id, st))
 			return 0;
-		if (!HBUF_PUTSL(ob, "\">"))
+		if (!HBUF_PUTSL(ob, "\""))
 			return 0;
-	} else {
-		if (!hbuf_printf(ob, "<h%zu>", level))
+	} else if (st->flags & LOWDOWN_HTML_HEAD_IDS) {
+		if (!HBUF_PUTSL(ob, " id=\""))
+			return 0;
+		buf = hbuf_id(NULL, n, &st->headers_used);
+		if (buf == NULL)
+			return 0;
+		if (!hbuf_putb(ob, buf))
+			return 0;
+		if (!HBUF_PUTSL(ob, "\""))
 			return 0;
 	}
 
+	/* Optional header class. */
+
+	if (n->rndr_header.attr_cls.size) {
+		if (!HBUF_PUTSL(ob, " class=\""))
+			return 0;
+	    	if (!escape_attr(ob, &n->rndr_header.attr_cls))
+			return 0;
+		if (!HBUF_PUTSL(ob, "\""))
+			return 0;
+	}
+
+	if (!HBUF_PUTSL(ob, ">"))
+		return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
 	return hbuf_printf(ob, "</h%zu>\n", level);
@@ -838,28 +809,9 @@ rndr_normal_text(struct lowdown_buf *ob,
 	return escape_htmlb(ob, &param->text, st);
 }
 
-static int
-rndr_footnotes(struct lowdown_buf *ob,
-	const struct lowdown_buf *content)
-{
-
-	if (ob->size && !hbuf_putc(ob, '\n'))
-		return 0;
-	if (!HBUF_PUTSL(ob, "<div class=\"footnotes\">\n"))
-		return 0;
-	if (!hbuf_puts(ob, "<hr/>\n"))
-		return 0;
-	if (!HBUF_PUTSL(ob, "<ol>\n"))
-		return 0;
-	if (!hbuf_putb(ob, content))
-		return 0;
-	return HBUF_PUTSL(ob, "\n</ol>\n</div>\n");
-}
-
 static int
 rndr_footnote_def(struct lowdown_buf *ob,
-	const struct lowdown_buf *content, 
-	const struct rndr_footnote_def *param)
+	const struct lowdown_buf *content, size_t num)
 {
 	size_t	i = 0;
 	int	pfound = 0;
@@ -881,7 +833,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 		break;
 	}
 
-	if (!hbuf_printf(ob, "\n<li id=\"fn%zu\">\n", param->num))
+	if (!hbuf_printf(ob, "\n<li id=\"fn%zu\">\n", num))
 		return 0;
 
 	if (pfound) {
@@ -890,7 +842,7 @@ rndr_footnote_def(struct lowdown_buf *ob,
 		if (!hbuf_printf(ob, "&#160;"
 		    "<a href=\"#fnref%zu\" rev=\"footnote\">"
 		    "&#8617;"
-		    "</a>", param->num))
+		    "</a>", num))
 			return 0;
 		if (!hbuf_put(ob, 
 		    content->data + i, content->size - i))
@@ -905,14 +857,29 @@ rndr_footnote_def(struct lowdown_buf *ob,
 
 static int
 rndr_footnote_ref(struct lowdown_buf *ob,
-	const struct rndr_footnote_ref *param)
+	const struct lowdown_buf *content, struct html *st)
 {
+	void	*pp;
+	size_t	 num = st->footsz + 1;
+
+	/*
+	 * Keep a reference to this footnote definition, as we're going
+	 * to print it out at the end of the document.  For now,
+	 * suppress printing of the content.
+	 */
+
+	pp = recallocarray(st->foots, st->footsz,
+		st->footsz + 1, sizeof(struct lowdown_buf *));
+	if (pp == NULL)
+		return 0;
+	st->foots = pp;
+	if ((st->foots[st->footsz++] = hbuf_dup(content)) == NULL)
+		return 0;
 
 	return hbuf_printf(ob, 
 		"<sup id=\"fnref%zu\">"
 		"<a href=\"#fn%zu\" rel=\"footnote\">"
-		"%zu</a></sup>", 
-		param->num, param->num, param->num);
+		"%zu</a></sup>", num, num, num);
 }
 
 static int
@@ -937,10 +904,28 @@ rndr_math(struct lowdown_buf *ob,
 static int
 rndr_doc_footer(struct lowdown_buf *ob, const struct html *st)
 {
+	size_t	 i;
 
-	if (st->flags & LOWDOWN_STANDALONE)
-		return HBUF_PUTSL(ob, "</body>\n");
-	return 1;
+	/*
+	 * Start by flushing out our footnotes.  Footnotes are "sparse"
+	 * in that we may not have them all defined (?).
+	 */
+
+	if (st->footsz > 0) {
+		if (ob->size && !hbuf_putc(ob, '\n'))
+			return 0;
+		if (!HBUF_PUTSL(ob,
+		    "<div class=\"footnotes\">\n<hr/>\n<ol>\n"))
+			return 0;
+		for (i = 0; i < st->footsz; i++)
+			if (!rndr_footnote_def(ob, st->foots[i], i + 1))
+				return 0;
+		if (!HBUF_PUTSL(ob, "\n</ol>\n</div>\n"))
+			return 0;
+	}
+
+	return (st->flags & LOWDOWN_STANDALONE) ?
+		HBUF_PUTSL(ob, "</body>\n") : 1;
 }
 
 static int
@@ -954,6 +939,8 @@ rndr_root(struct lowdown_buf *ob,
 		    return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
+	if (!rndr_doc_footer(ob, st))
+		return 0;
 	if (st->flags & LOWDOWN_STANDALONE)
 		return HBUF_PUTSL(ob, "</html>\n");
 	return 1;
@@ -1184,10 +1171,15 @@ rndr(struct lowdown_buf *ob,
 		if (!rndr(tmp, mq, st, child))
 			goto out;
 
-	if (n->chng == LOWDOWN_CHNG_INSERT && 
+	/*
+	 * If we're in the doc header, don't emit any insert or delete,
+	 * as HTML doesn't allow them.
+	 */
+
+	if (n->chng == LOWDOWN_CHNG_INSERT && n->type != LOWDOWN_META &&
 	    !HBUF_PUTSL(ob, "<ins>"))
 		goto out;
-	if (n->chng == LOWDOWN_CHNG_DELETE && 
+	if (n->chng == LOWDOWN_CHNG_DELETE && n->type != LOWDOWN_META &&
 	   !HBUF_PUTSL(ob, "<del>"))
 		goto out;
 
@@ -1218,11 +1210,8 @@ rndr(struct lowdown_buf *ob,
 		if (n->chng != LOWDOWN_CHNG_DELETE)
 			rc = rndr_meta(ob, tmp, mq, n, st);
 		break;
-	case LOWDOWN_DOC_FOOTER:
-		rc = rndr_doc_footer(ob, st);
-		break;
 	case LOWDOWN_HEADER:
-		rc = rndr_header(ob, tmp, &n->rndr_header, st);
+		rc = rndr_header(ob, tmp, n, st);
 		break;
 	case LOWDOWN_HRULE:
 		rc = rndr_hrule(ob);
@@ -1251,12 +1240,6 @@ rndr(struct lowdown_buf *ob,
 	case LOWDOWN_TABLE_CELL:
 		rc = rndr_tablecell(ob, tmp, &n->rndr_table_cell);
 		break;
-	case LOWDOWN_FOOTNOTES_BLOCK:
-		rc = rndr_footnotes(ob, tmp);
-		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-		rc = rndr_footnote_def(ob, tmp, &n->rndr_footnote_def);
-		break;
 	case LOWDOWN_BLOCKHTML:
 		rc = rndr_raw_block(ob, &n->rndr_blockhtml, st);
 		break;
@@ -1293,8 +1276,8 @@ rndr(struct lowdown_buf *ob,
 	case LOWDOWN_SUPERSCRIPT:
 		rc = rndr_superscript(ob, tmp);
 		break;
-	case LOWDOWN_FOOTNOTE_REF:
-		rc = rndr_footnote_ref(ob, &n->rndr_footnote_ref);
+	case LOWDOWN_FOOTNOTE:
+		rc = rndr_footnote_ref(ob, tmp, st);
 		break;
 	case LOWDOWN_MATH_BLOCK:
 		rc = rndr_math(ob, &n->rndr_math, st);
@@ -1330,10 +1313,17 @@ rndr(struct lowdown_buf *ob,
 	if (!rc)
 		goto out;
 
-	if (n->chng == LOWDOWN_CHNG_INSERT && 
+	/*
+	 * If we're in the doc header, don't emit any insert or delete,
+	 * as HTML doesn't allow them.
+	 */
+
+	if (n->chng == LOWDOWN_CHNG_INSERT && n->type != LOWDOWN_META &&
+	    n->parent != NULL &&
+	    n->parent->type != LOWDOWN_DOC_HEADER &&
 	    !HBUF_PUTSL(ob, "</ins>"))
 		goto out;
-	if (n->chng == LOWDOWN_CHNG_DELETE &&
+	if (n->chng == LOWDOWN_CHNG_DELETE && n->type != LOWDOWN_META &&
 	    !HBUF_PUTSL(ob, "</del>"))
 		goto out;
 
@@ -1350,13 +1340,22 @@ lowdown_html_rndr(struct lowdown_buf *ob,
 	struct html		*st = arg;
 	struct lowdown_metaq	 metaq;
 	int			 rc;
+	size_t			 i;
 
+	TAILQ_INIT(&st->headers_used);
 	TAILQ_INIT(&metaq);
 	st->headers_offs = 1;
 
 	rc = rndr(ob, &metaq, st, n);
 
+	for (i = 0; i < st->footsz; i++)
+		hbuf_free(st->foots[i]);
+
+	free(st->foots);
+	st->footsz = 0;
+	st->foots = NULL;
 	lowdown_metaq_free(&metaq);
+	hentryq_clear(&st->headers_used);
 	return rc;
 }
 
@@ -1368,7 +1367,6 @@ lowdown_html_new(const struct lowdown_opts *opts)
 	if ((p = calloc(1, sizeof(struct html))) == NULL)
 		return NULL;
 
-	TAILQ_INIT(&p->headers_used);
 	p->flags = opts == NULL ? 0 : opts->oflags;
 	return p;
 }
@@ -1376,17 +1374,6 @@ lowdown_html_new(const struct lowdown_opts *opts)
 void
 lowdown_html_free(void *arg)
 {
-	struct html	*st = arg;
-	struct hentry	*hentry;
 
-	if (st == NULL)
-		return;
-
-	while ((hentry = TAILQ_FIRST(&st->headers_used)) != NULL) {
-		TAILQ_REMOVE(&st->headers_used, hentry, entries);
-		free(hentry->str);
-		free(hentry);
-	}
-
-	free(st);
+	free(arg);
 }
