@@ -20,6 +20,7 @@
 # include <sys/queue.h>
 #endif
 
+#include <assert.h>
 #include <ctype.h>
 #include <inttypes.h>
 #include <stdint.h>
@@ -33,7 +34,9 @@
 
 struct latex {
 	unsigned int	oflags; /* same as in lowdown_opts */
+	struct hentryq	headers_used; /* headers we've seen */
 	ssize_t		headers_offs; /* header offset */
+	size_t		footsz; /* current footnote */
 };
 
 /*
@@ -116,6 +119,8 @@ rndr_entity(struct lowdown_buf *ob,
 	if (tex == NULL)
 		return rndr_escape(ob, &param->text);
 
+	if ((texflags & TEX_ENT_MATH) && (texflags & TEX_ENT_ASCII))
+		return hbuf_printf(ob, "$\\mathrm{%s}$", tex);
 	if (texflags & TEX_ENT_ASCII)
 		return hbuf_puts(ob, tex);
 	if (texflags & TEX_ENT_MATH)
@@ -261,18 +266,38 @@ rndr_linebreak(struct lowdown_buf *ob)
 }
 
 static int
-rndr_header(struct lowdown_buf *ob,
-	const struct lowdown_buf *content,
-	const struct rndr_header *param,
-	const struct latex *st)
+rndr_header(struct lowdown_buf *ob, const struct lowdown_buf *content,
+	const struct lowdown_node *n, struct latex *st)
 {
-	const char	*type;
-	ssize_t		 level;
+	const char			*type;
+	ssize_t				 level;
+	struct lowdown_buf		*buf = NULL;
+	const struct lowdown_buf	*id;
+	int				 rc = 0;
+
+	if (n->rndr_header.attr_id.size) {
+		if ((buf = hbuf_new(32)) == NULL)
+			goto out;
+		if (!rndr_escape(buf, &n->rndr_header.attr_id))
+			goto out;
+		id = buf;
+	} else {
+		id = hbuf_id(NULL, n, &st->headers_used);
+		if (id == NULL)
+			goto out;
+	}
 
 	if (ob->size && !HBUF_PUTSL(ob, "\n"))
-		return 0;
+		goto out;
 
-	level = (ssize_t)param->level + st->headers_offs;
+	if (!HBUF_PUTSL(ob, "\\hypertarget{"))
+		goto out;
+	if (!hbuf_putb(ob, id))
+		goto out;
+	if (!HBUF_PUTSL(ob, "}{%\n"))
+		goto out;
+
+	level = (ssize_t)n->rndr_header.level + st->headers_offs;
 	if (level < 1)
 		level = 1;
 
@@ -295,15 +320,24 @@ rndr_header(struct lowdown_buf *ob,
 	}
 
 	if (!hbuf_puts(ob, type))
-		return 0;
+		goto out;
 	if (!(st->oflags & LOWDOWN_LATEX_NUMBERED) &&
   	    !HBUF_PUTSL(ob, "*"))
-		return 0;
+		goto out;
 	if (!HBUF_PUTSL(ob, "{"))
-		return 0;
+		goto out;
 	if (!hbuf_putb(ob, content))
-		return 0;
-	return HBUF_PUTSL(ob, "}\n");
+		goto out;
+	if (!HBUF_PUTSL(ob, "}\\label{"))
+		goto out;
+	if (!hbuf_putb(ob, id))
+		goto out;
+	if (!HBUF_PUTSL(ob, "}}\n"))
+		goto out;
+	rc = 1;
+out:
+	hbuf_free(buf);
+	return rc;
 }
 
 static int
@@ -311,15 +345,36 @@ rndr_link(struct lowdown_buf *ob,
 	const struct lowdown_buf *content,
 	const struct rndr_link *param)
 {
+	int	loc;
 
-	if (!HBUF_PUTSL(ob, "\\href{"))
+	loc = param->link.size > 0 &&
+		param->link.data[0] == '#';
+
+	if (param->attr_id.size > 0) {
+		if (!HBUF_PUTSL(ob, "\\hypertarget{"))
+			return 0;
+		if (!hbuf_putb(ob, &param->attr_id))
+			return 0;
+		if (!HBUF_PUTSL(ob, "}{%\n"))
+			return 0;
+	}
+
+	if (loc && !HBUF_PUTSL(ob, "\\hyperlink{"))
 		return 0;
-	if (!rndr_escape(ob, &param->link))
+	else if (!loc && !HBUF_PUTSL(ob, "\\href{"))
+		return 0;
+
+	if (loc && !rndr_escape_text
+	    (ob, &param->link.data[1], param->link.size - 1))
+		return 0;
+	else if (!loc && !rndr_escape(ob, &param->link))
 		return 0;
 	if (!HBUF_PUTSL(ob, "}{"))
 		return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
+	if (param->attr_id.size > 0 && !HBUF_PUTSL(ob, "}"))
+		return 0;
 	return HBUF_PUTSL(ob, "}");
 }
 
@@ -465,7 +520,8 @@ rndr_image(struct lowdown_buf *ob,
 	if (!HBUF_PUTSL(ob, "\\includegraphics["))
 		return 0;
 	if (param->attr_width.size || param->attr_height.size) {
-		if (param->attr_width.size) {
+		if (param->attr_width.size &&
+		    param->attr_width.size < sizeof(dimbuf) - 1) {
 			memset(dimbuf, 0, sizeof(dimbuf));
 			memcpy(dimbuf, param->attr_width.data, 
 				param->attr_width.size);
@@ -483,7 +539,8 @@ rndr_image(struct lowdown_buf *ob,
 					return 0;
 			}
 		}
-		if (param->attr_height.size) {
+		if (param->attr_height.size &&
+		    param->attr_height.size < sizeof(dimbuf) - 1) {
 			if (param->attr_width.size && 
 			    !HBUF_PUTSL(ob, ", "))
 				return 0;
@@ -542,9 +599,7 @@ rndr_table(struct lowdown_buf *ob,
 		return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
-	if (!HBUF_PUTSL(ob, "\\end{tabular}\n"))
-		return 0;
-	return HBUF_PUTSL(ob, "\\end{center}\n");
+	return HBUF_PUTSL(ob, "\\end{longtable}\n");
 }
 
 static int
@@ -553,17 +608,23 @@ rndr_table_header(struct lowdown_buf *ob,
 	const struct rndr_table_header *param)
 {
 	size_t	 i;
+	char	 align;
+	int	 fl;
 
-	if (!HBUF_PUTSL(ob, "\\begin{center}"))
-		return 0;
-	if (!HBUF_PUTSL(ob, "\\begin{tabular}{ "))
+	if (!HBUF_PUTSL(ob, "\\begin{longtable}[]{"))
 		return 0;
 
-	/* FIXME: alignment */
-
-	for (i = 0; i < param->columns; i++)
-		if (!HBUF_PUTSL(ob, "c "))
+	for (i = 0; i < param->columns; i++) {
+		fl = param->flags[i] & HTBL_FL_ALIGNMASK;
+		if (fl == HTBL_FL_ALIGN_CENTER)
+			align = 'c';
+		else if (fl == HTBL_FL_ALIGN_RIGHT)
+			align = 'r';
+		else
+			align = 'l';
+		if (!hbuf_putc(ob, align))
 			return 0;
+	}
 	if (!HBUF_PUTSL(ob, "}\n"))
 		return 0;
 	return hbuf_putb(ob, content);
@@ -603,37 +664,17 @@ rndr_normal_text(struct lowdown_buf *ob,
 }
 
 static int
-rndr_footnote_def(struct lowdown_buf *ob,
-	const struct lowdown_buf *content, 
-	const struct lowdown_node *n,
-	const struct rndr_footnote_def *param)
+rndr_footnote_ref(struct lowdown_buf *ob,
+	const struct lowdown_buf *content, struct latex *st)
 {
 
-	if (!hbuf_printf(ob, "\\footnotetext[%zu]{", param->num))
-		return 0;
-	if (n->chng == LOWDOWN_CHNG_INSERT &&
-	    !HBUF_PUTSL(ob, "\\textcolor{blue}{"))
-		return 0;
-	if (n->chng == LOWDOWN_CHNG_DELETE &&
-	    !HBUF_PUTSL(ob, "\\textcolor{red}{"))
+	if (!hbuf_printf(ob, "\\footnote[%zu]{", ++st->footsz))
 		return 0;
 	if (!hbuf_putb(ob, content))
 		return 0;
-	if ((n->chng == LOWDOWN_CHNG_INSERT ||
-	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    !HBUF_PUTSL(ob, "}"))
-		return 0;
 	return HBUF_PUTSL(ob, "}\n");
 }
 
-static int
-rndr_footnote_ref(struct lowdown_buf *ob,
-	const struct rndr_footnote_ref *param)
-{
-
-	return hbuf_printf(ob, "\\footnotemark[%zu]", param->num);
-}
-
 static int
 rndr_math(struct lowdown_buf *ob,
 	const struct rndr_math *param)
@@ -675,16 +716,27 @@ rndr_doc_header(struct lowdown_buf *ob,
 		return 1;
 
 	if (!HBUF_PUTSL(ob, 
+	    "% Options for packages loaded elsewhere\n"
+	    "\\PassOptionsToPackage{unicode}{hyperref}\n"
+	    "\\PassOptionsToPackage{hyphens}{url}\n"
+	    "%\n"
 	    "\\documentclass[11pt,a4paper]{article}\n"
+	    "\\usepackage{amsmath,amssymb}\n"
+	    "\\usepackage{lmodern}\n"
+	    "\\usepackage{iftex}\n"
+	    "\\ifPDFTeX\n"
+	    "  \\usepackage[T1]{fontenc}\n"
+	    "  \\usepackage[utf8]{inputenc}\n"
+	    "  \\usepackage{textcomp} % provide euro and other symbols\n"
+	    "\\else % if luatex or xetex\n"
+	    "  \\usepackage{unicode-math}\n"
+	    "  \\defaultfontfeatures{Scale=MatchLowercase}\n"
+	    "  \\defaultfontfeatures[\\rmfamily]{Ligatures=TeX,Scale=1}\n"
+	    "\\fi\n"
 	    "\\usepackage{xcolor}\n"
 	    "\\usepackage{graphicx}\n"
-	    "\\usepackage[utf8]{inputenc}\n"
-	    "\\usepackage[T1]{fontenc}\n"
-	    "\\usepackage{textcomp}\n"
-	    "\\usepackage{lmodern}\n"
+	    "\\usepackage{longtable}\n"
 	    "\\usepackage{hyperref}\n"
-	    "\\usepackage{amsmath}\n"
-	    "\\usepackage{amssymb}\n"
 	    "\\begin{document}\n"))
 		return 0;
 
@@ -775,7 +827,7 @@ rndr(struct lowdown_buf *ob,
 	struct lowdown_buf		*tmp;
 	struct latex			*st = arg;
 	const struct lowdown_node	*child;
-	int				 ret = 0, rc = 1;
+	int				 ret = 0;
 
 	if ((tmp = hbuf_new(64)) == NULL)
 		return 0;
@@ -790,124 +842,148 @@ rndr(struct lowdown_buf *ob,
 	 */
 
 	if (n->chng == LOWDOWN_CHNG_INSERT && 
-	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !HBUF_PUTSL(ob, "{\\color{blue} "))
 		goto out;
 	if (n->chng == LOWDOWN_CHNG_DELETE &&
-	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !HBUF_PUTSL(ob, "{\\color{red} "))
 		goto out;
 
 	switch (n->type) {
 	case LOWDOWN_BLOCKCODE:
-		rc = rndr_blockcode(ob, &n->rndr_blockcode);
+		if (!rndr_blockcode(ob, &n->rndr_blockcode))
+			return 0;
 		break;
 	case LOWDOWN_BLOCKQUOTE:
-		rc = rndr_blockquote(ob, tmp);
+		if (!rndr_blockquote(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_DEFINITION:
-		rc = rndr_definition(ob, tmp);
+		if (!rndr_definition(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_DEFINITION_TITLE:
-		rc = rndr_definition_title(ob, tmp);
+		if (!rndr_definition_title(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_DOC_HEADER:
-		rc = rndr_doc_header(ob, mq, st);
+		if (!rndr_doc_header(ob, mq, st))
+			return 0;
 		break;
 	case LOWDOWN_META:
-		if (n->chng != LOWDOWN_CHNG_DELETE)
-			rc = rndr_meta(ob, tmp, mq, n, st);
-		break;
-	case LOWDOWN_DOC_FOOTER:
-		rc = rndr_doc_footer(ob, st);
+		if (n->chng != LOWDOWN_CHNG_DELETE &&
+		    !rndr_meta(ob, tmp, mq, n, st))
+			return 0;
 		break;
 	case LOWDOWN_HEADER:
-		rc = rndr_header(ob, tmp, &n->rndr_header, st);
+		if (!rndr_header(ob, tmp, n, st))
+			return 0;
 		break;
 	case LOWDOWN_HRULE:
-		rc = rndr_hrule(ob);
+		if (!rndr_hrule(ob))
+			return 0;
 		break;
 	case LOWDOWN_LIST:
-		rc = rndr_list(ob, tmp, &n->rndr_list);
+		if (!rndr_list(ob, tmp, &n->rndr_list))
+			return 0;
 		break;
 	case LOWDOWN_LISTITEM:
-		rc = rndr_listitem(ob, tmp, &n->rndr_listitem);
+		if (!rndr_listitem(ob, tmp, &n->rndr_listitem))
+			return 0;
 		break;
 	case LOWDOWN_PARAGRAPH:
-		rc = rndr_paragraph(ob, tmp);
+		if (!rndr_paragraph(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_TABLE_BLOCK:
-		rc = rndr_table(ob, tmp);
+		if (!rndr_table(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_TABLE_HEADER:
-		rc = rndr_table_header(ob, tmp, &n->rndr_table_header);
+		if (!rndr_table_header(ob, tmp, &n->rndr_table_header))
+			return 0;
 		break;
 	case LOWDOWN_TABLE_CELL:
-		rc = rndr_tablecell(ob, tmp, &n->rndr_table_cell);
-		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-		rc = rndr_footnote_def
-			(ob, tmp, n, &n->rndr_footnote_def);
+		if (!rndr_tablecell(ob, tmp, &n->rndr_table_cell))
+			return 0;
 		break;
 	case LOWDOWN_BLOCKHTML:
-		rc = rndr_raw_block(ob, &n->rndr_blockhtml, st);
+		if (!rndr_raw_block(ob, &n->rndr_blockhtml, st))
+			return 0;
 		break;
 	case LOWDOWN_LINK_AUTO:
-		rc = rndr_autolink(ob, &n->rndr_autolink);
+		if (!rndr_autolink(ob, &n->rndr_autolink))
+			return 0;
 		break;
 	case LOWDOWN_CODESPAN:
-		rc = rndr_codespan(ob, &n->rndr_codespan);
+		if (!rndr_codespan(ob, &n->rndr_codespan))
+			return 0;
 		break;
 	case LOWDOWN_DOUBLE_EMPHASIS:
-		rc = rndr_double_emphasis(ob, tmp);
+		if (!rndr_double_emphasis(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_EMPHASIS:
-		rc = rndr_emphasis(ob, tmp);
+		if (!rndr_emphasis(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_HIGHLIGHT:
-		rc = rndr_highlight(ob, tmp);
+		if (!rndr_highlight(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_IMAGE:
-		rc = rndr_image(ob, &n->rndr_image);
+		if (!rndr_image(ob, &n->rndr_image))
+			return 0;
 		break;
 	case LOWDOWN_LINEBREAK:
-		rc = rndr_linebreak(ob);
+		if (!rndr_linebreak(ob))
+			return 0;
 		break;
 	case LOWDOWN_LINK:
-		rc = rndr_link(ob, tmp, &n->rndr_link);
+		if (!rndr_link(ob, tmp, &n->rndr_link))
+			return 0;
 		break;
 	case LOWDOWN_TRIPLE_EMPHASIS:
-		rc = rndr_triple_emphasis(ob, tmp);
+		if (!rndr_triple_emphasis(ob, tmp))
+			return 0;
 		break;
 	case LOWDOWN_SUPERSCRIPT:
-		rc = rndr_superscript(ob, tmp);
+		if (!rndr_superscript(ob, tmp))
+			return 0;
 		break;
-	case LOWDOWN_FOOTNOTE_REF:
-		rc = rndr_footnote_ref(ob, &n->rndr_footnote_ref);
+	case LOWDOWN_FOOTNOTE:
+		if (!rndr_footnote_ref(ob, tmp, st))
+			return 0;
 		break;
 	case LOWDOWN_MATH_BLOCK:
-		rc = rndr_math(ob, &n->rndr_math);
+		if (!rndr_math(ob, &n->rndr_math))
+			return 0;
 		break;
 	case LOWDOWN_RAW_HTML:
-		rc = rndr_raw_html(ob, &n->rndr_raw_html, st);
+		if (!rndr_raw_html(ob, &n->rndr_raw_html, st))
+			return 0;
 		break;
 	case LOWDOWN_NORMAL_TEXT:
-		rc = rndr_normal_text(ob, &n->rndr_normal_text);
+		if (!rndr_normal_text(ob, &n->rndr_normal_text))
+			return 0;
 		break;
 	case LOWDOWN_ENTITY:
-		rc = rndr_entity(ob, &n->rndr_entity);
+		if (!rndr_entity(ob, &n->rndr_entity))
+			return 0;
+		break;
+	case LOWDOWN_ROOT:
+		if (!hbuf_putb(ob, tmp))
+			return 0;
+		if (!rndr_doc_footer(ob, st))
+			return 0;
 		break;
 	default:
-		rc = hbuf_putb(ob, tmp);
+		if (!hbuf_putb(ob, tmp))
+			return 0;
 		break;
 	}
-	if (!rc)
-		goto out;
 
 	if ((n->chng == LOWDOWN_CHNG_INSERT ||
-	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    n->type != LOWDOWN_FOOTNOTE_DEF &&
-	    !HBUF_PUTSL(ob, "}"))
+	     n->chng == LOWDOWN_CHNG_DELETE) && !HBUF_PUTSL(ob, "}"))
 		goto out;
 
 	ret = 1;
@@ -924,12 +1000,15 @@ lowdown_latex_rndr(struct lowdown_buf *ob,
 	struct lowdown_metaq	 metaq;
 	int			 rc;
 
+	TAILQ_INIT(&st->headers_used);
 	TAILQ_INIT(&metaq);
 	st->headers_offs = 1;
+	st->footsz = 0;
 
 	rc = rndr(ob, &metaq, st, n);
 
 	lowdown_metaq_free(&metaq);
+	hentryq_clear(&st->headers_used);
 	return rc;
 }
 
