@@ -72,8 +72,6 @@ enum	lowdown_rndrt {
 	LOWDOWN_TABLE_BODY,
 	LOWDOWN_TABLE_ROW,
 	LOWDOWN_TABLE_CELL,
-	LOWDOWN_FOOTNOTES_BLOCK,
-	LOWDOWN_FOOTNOTE_DEF,
 	LOWDOWN_BLOCKHTML,
 	LOWDOWN_LINK_AUTO,
 	LOWDOWN_CODESPAN,
@@ -86,14 +84,13 @@ enum	lowdown_rndrt {
 	LOWDOWN_TRIPLE_EMPHASIS,
 	LOWDOWN_STRIKETHROUGH,
 	LOWDOWN_SUPERSCRIPT,
-	LOWDOWN_FOOTNOTE_REF,
+	LOWDOWN_FOOTNOTE,
 	LOWDOWN_MATH_BLOCK,
 	LOWDOWN_RAW_HTML,
 	LOWDOWN_ENTITY,
 	LOWDOWN_NORMAL_TEXT,
 	LOWDOWN_DOC_HEADER,
 	LOWDOWN_META,
-	LOWDOWN_DOC_FOOTER,
 	LOWDOWN__MAX
 };
 
@@ -117,8 +114,8 @@ enum 	htbl_flags {
 
 enum 	halink_type {
 	HALINK_NONE, /* used internally when it is not an autolink */
-	HALINK_NORMAL, /* normal http/http/ftp/mailto/etc link */
-	HALINK_EMAIL /* e-mail link without explit mailto: */
+	HALINK_NORMAL,
+	HALINK_EMAIL
 };
 
 enum	hlist_fl {
@@ -153,25 +150,25 @@ struct	rndr_meta {
 };
 
 struct	rndr_paragraph {
-	size_t lines; /* input lines */
-	int beoln; /* ends on blank line */
+	size_t lines;
+	int beoln;
 };
 
 struct	rndr_normal_text {
-	struct lowdown_buf text; /* basic text */
+	struct lowdown_buf text;
 };
 
 struct	rndr_entity {
-	struct lowdown_buf text; /* entity text */
+	struct lowdown_buf text;
 };
 
 struct	rndr_autolink {
-	struct lowdown_buf link; /* link address */
-	enum halink_type type; /* type of link */
+	struct lowdown_buf link;
+	enum halink_type type;
 };
 
 struct	rndr_raw_html {
-	struct lowdown_buf text; /* raw html buffer */
+	struct lowdown_buf text;
 };
 
 struct	rndr_link {
@@ -182,8 +179,8 @@ struct	rndr_link {
 };
 
 struct	rndr_blockcode {
-	struct lowdown_buf text; /* raw code buffer */
-	struct lowdown_buf lang; /* fence language */
+	struct lowdown_buf text;
+	struct lowdown_buf lang;
 };
 
 struct	rndr_definition {
@@ -191,7 +188,7 @@ struct	rndr_definition {
 };
 
 struct	rndr_codespan {
-	struct lowdown_buf text; /* raw code buffer */
+	struct lowdown_buf text;
 };
 
 struct	rndr_table{
@@ -209,17 +206,6 @@ struct	rndr_table_cell {
 	size_t columns;
 };
 
-struct	rndr_footnote_def {
-	size_t num;
-	struct lowdown_buf key;
-};
-
-struct	rndr_footnote_ref {
-	size_t num;
-	struct lowdown_buf def;
-	struct lowdown_buf key;
-};
-
 struct	rndr_blockhtml {
 	struct lowdown_buf text;
 };
@@ -235,7 +221,9 @@ struct	rndr_listitem {
 };
 
 struct	rndr_header{
-	size_t level; /* hN level */
+	size_t level;
+	struct lowdown_buf attr_cls;
+	struct lowdown_buf attr_id;
 };
 
 struct	rndr_image {
@@ -250,7 +238,7 @@ struct	rndr_image {
 };
 
 struct rndr_math {
-	struct lowdown_buf text; /* equation (opaque) */
+	struct lowdown_buf text;
 	int blockmode;
 };
 
@@ -279,8 +267,6 @@ struct	lowdown_node {
 		struct rndr_table rndr_table; 
 		struct rndr_table_header rndr_table_header; 
 		struct rndr_table_cell rndr_table_cell; 
-		struct rndr_footnote_def rndr_footnote_def;
-		struct rndr_footnote_ref rndr_footnote_ref;
 		struct rndr_image rndr_image;
 		struct rndr_math rndr_math;
 		struct rndr_blockhtml rndr_blockhtml;
@@ -290,15 +276,19 @@ struct	lowdown_node {
 	TAILQ_ENTRY(lowdown_node) entries;
 };
 
-/*
- * These options contain everything needed to parse and render content.
- */
+struct	lowdown_opts_odt {
+	const char		*sty;
+};
+
 struct	lowdown_opts {
 	enum lowdown_type	  type;
-	size_t			  maxdepth; /* max parse tree depth */
-	size_t			  cols; /* -Tterm width */
-	size_t			  hmargin; /* -Tterm left margin */
-	size_t			  vmargin; /* -Tterm top/bot margin */
+	union {
+		struct lowdown_opts_odt odt;
+	};
+	size_t			  maxdepth;
+	size_t			  cols;
+	size_t			  hmargin;
+	size_t			  vmargin;
 	unsigned int		  feat;
 #define LOWDOWN_TABLES		  0x01
 #define LOWDOWN_FENCED		  0x02
