@@ -41,16 +41,20 @@ enum	nfont {
 };
 
 struct 	nroff {
-	int 		 man; /* whether man(7) */
-	int		 post_para; /* for choosing PP/LP */
-	unsigned int 	 flags; /* output flags */
-	ssize_t		 headers_offs; /* header offset */
-	enum nfont	 fonts[NFONT__MAX]; /* see bqueue_font() */
+	struct hentryq	 	  headers_used; /* headers we've seen */
+	int			  man; /* whether man(7) */
+	int			  post_para; /* for choosing PP/LP */
+	unsigned int		  flags; /* output flags */
+	ssize_t			  headers_offs; /* header offset */
+	enum nfont		  fonts[NFONT__MAX]; /* see bqueue_font() */
+	struct bnodeq		**foots; /* footnotes */
+	size_t			  footsz; /* footnote size */
 };
 
 enum	bscope {
 	BSCOPE_BLOCK = 0,
-	BSCOPE_SPAN = 1,
+	BSCOPE_SPAN,
+	BSCOPE_PDFHREF,
 	BSCOPE_LITERAL,
 	BSCOPE_FONT,
 	BSCOPE_COLOUR
@@ -59,14 +63,11 @@ enum	bscope {
 /*
  * Instead of writing directly into the output buffer, we write
  * temporarily into bnodes, which are converted into output.  These
- * nodes are aware of whether they need surrounding newlines.  This way,
- * we have much more control over where to put newlines, which before
- * was haphazard at best.
+ * nodes are aware of whether they need surrounding newlines.
  */
 struct	bnode {
 	char				*nbuf; /* (safe) 1st data */
 	char				*buf; /* (unsafe) 2nd data */
-	size_t				 bufchop; /* strip from buf */
 	char				*nargs; /* (safe) 1st args */
 	char				*args; /* (unsafe) 2nd args */
 	int				 close; /* BNODE_COLOUR/FONT */
@@ -95,11 +96,26 @@ static int
 hesc_nroff(struct lowdown_buf *ob, const char *data, 
 	size_t size, int oneline, int literal, int esc)
 {
-	size_t	 i = 0;
+	size_t	 	i = 0;
+	unsigned char	ch;
 
 	if (size == 0)
 		return 1;
-	if (!esc)
+
+	if (!esc && oneline) {
+		assert(!literal);
+		for (i = 0; i < size; i++) {
+			ch = data[i] == '\n' ? ' ' : data[i];
+			if (!hbuf_putc(ob, ch))
+				return 0;
+			if (ch != ' ')
+				continue;
+			while (i < size && isspace((unsigned char)data[i]))
+				i++;
+			i--;
+		}
+		return 1;
+	} else if (!esc)
 		return hbuf_put(ob, data, size);
 
 	/* Strip leading whitespace. */
@@ -261,13 +277,13 @@ bqueue_font(const struct nroff *st, struct bnodeq *bq, int close)
 }
 
 static struct bnode *
-bqueue_span(struct bnodeq *bq, const char *text)
+bqueue_node(struct bnodeq *bq, enum bscope scope, const char *text)
 {
 	struct bnode	*bn;
 
 	if ((bn = calloc(1, sizeof(struct bnode))) == NULL)
 		return NULL;
-	bn->scope = BSCOPE_SPAN;
+	bn->scope = scope;
 	if (text != NULL && (bn->nbuf = strdup(text)) == NULL) {
 		free(bn);
 		return NULL;
@@ -277,19 +293,24 @@ bqueue_span(struct bnodeq *bq, const char *text)
 }
 
 static struct bnode *
-bqueue_block(struct bnodeq *bq, const char *macro)
+bqueue_span(struct bnodeq *bq, const char *text)
 {
-	struct bnode	*bn;
 
-	if ((bn = calloc(1, sizeof(struct bnode))) == NULL)
-		return NULL;
-	bn->scope = BSCOPE_BLOCK;
-	if (macro != NULL && (bn->nbuf = strdup(macro)) == NULL) {
-		free(bn);
-		return NULL;
-	}
-	TAILQ_INSERT_TAIL(bq, bn, entries);
-	return bn;
+	return bqueue_node(bq, BSCOPE_SPAN, text);
+}
+
+static struct bnode *
+bqueue_block(struct bnodeq *bq, const char *text)
+{
+
+	return bqueue_node(bq, BSCOPE_BLOCK, text);
+}
+
+static struct bnode *
+bqueue_semiblock(struct bnodeq *bq, const char *text)
+{
+
+	return bqueue_node(bq, BSCOPE_PDFHREF, text);
 }
 
 static void
@@ -334,13 +355,19 @@ bqueue_strip_paras(struct bnodeq *bq)
 static int
 bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 {
-	const struct bnode	*bn, *chk;
+	const struct bnode	*bn, *chk, *next;
 	const char		*cp;
 	int		 	 nextblk;
 
 	TAILQ_FOREACH(bn, bq, entries) {
 		nextblk = 0;
 
+		if (bn->scope == BSCOPE_PDFHREF &&
+		    ob->size > 0 && 
+		    ob->data[ob->size - 1] != '\n' &&
+		    !hbuf_puts(ob, "\\c"))
+			return 0;
+
 		/* 
 		 * Block scopes start with a newline.
 		 * Also have colours use their own block, as otherwise
@@ -351,6 +378,7 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 		 */
 
 		if (bn->scope == BSCOPE_BLOCK ||
+		    bn->scope == BSCOPE_PDFHREF ||
 		    bn->scope == BSCOPE_COLOUR) {
 			if (ob->size > 0 && 
 			    ob->data[ob->size - 1] != '\n' &&
@@ -369,7 +397,8 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 				TAILQ_PREV(bn, bnodeq, entries) :
 				TAILQ_NEXT(bn, entries);
 			if (chk != NULL && 
-			    chk->scope == BSCOPE_BLOCK) {
+			    (chk->scope == BSCOPE_PDFHREF ||
+			     chk->scope == BSCOPE_BLOCK)) {
 				if (ob->size > 0 && 
 				    ob->data[ob->size - 1] != '\n' &&
 				    !hbuf_putc(ob, '\n'))
@@ -407,6 +436,18 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 			if (!hbuf_putc(ob, '\n'))
 				return 0;
 
+		/*
+		 * If we're a span, double-check to see whether we
+		 * should introduce a line with an escape.
+		 */
+
+		if (bn->scope == BSCOPE_SPAN &&
+		    bn->nbuf != NULL && ob->size > 0 &&
+		    ob->data[ob->size - 1] == '\n' &&
+		    (bn->nbuf[0] == '.' || bn->nbuf[0] == '\'') &&
+		    !HBUF_PUTSL(ob, "\\&"))
+			return 0;
+
 		/* Safe data need not be escaped. */
 
 		if (bn->nbuf != NULL && !hbuf_puts(ob, bn->nbuf))
@@ -420,18 +461,28 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 			    strlen(bn->buf), 0, 1, esc))
 				return 0;
 		} else if (bn->buf != NULL)
-			if (!hesc_nroff(ob, bn->buf + bn->bufchop, 
-			    strlen(bn->buf) - bn->bufchop, 0, 0, esc))
+			if (!hesc_nroff(ob, bn->buf,
+			    strlen(bn->buf), 0, 0, esc))
 				return 0;
 
+		if (bn->scope == BSCOPE_PDFHREF &&
+		    (next = TAILQ_NEXT(bn, entries)) != NULL &&
+		    next->scope == BSCOPE_SPAN &&
+		    next->buf != NULL &&
+		    next->buf[0] != ' ' &&
+		    next->buf[0] != '\n' &&
+		    !HBUF_PUTSL(ob, " -A \"\\c\""))
+			return 0;
+
 		/* 
-		 * Macro arguments follow after space.
-		 * These must all be printed on the same line.
+		 * Macro arguments follow after space.  For links, these
+		 * must all be printed on the same line.
 		 */
 
-		if (bn->nargs != NULL) {
+		if (bn->nargs != NULL &&
+		    (bn->scope == BSCOPE_BLOCK ||
+		     bn->scope == BSCOPE_PDFHREF)) {
 			assert(nextblk);
-			assert(bn->scope == BSCOPE_BLOCK);
 			if (!hbuf_putc(ob, ' '))
 				return 0;
 			for (cp = bn->nargs; *cp != '\0'; cp++)
@@ -442,7 +493,8 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 
 		if (bn->args != NULL) {
 			assert(nextblk);
-			assert(bn->scope == BSCOPE_BLOCK);
+			assert(bn->scope == BSCOPE_BLOCK ||
+				bn->scope == BSCOPE_PDFHREF);
 			if (!hbuf_putc(ob, ' '))
 				return 0;
 			if (!hesc_nroff(ob, bn->args,
@@ -461,29 +513,6 @@ bqueue_flush(struct lowdown_buf *ob, const struct bnodeq *bq, int esc)
 	return 1;
 }
 
-static int
-bqueue_to_nargs(struct bnode *bn, const struct bnodeq *bq, int quoted)
-{
-	struct lowdown_buf	*ob;
-	int			 rc = 0;
-
-	if ((ob = hbuf_new(32)) == NULL)
-		goto out;
-	if (quoted && !hbuf_putc(ob, '"'))
-		goto out;
-	if (!bqueue_flush(ob, bq, 1))
-		goto out;
-	if (quoted && !hbuf_putc(ob, '"'))
-		goto out;
-	assert(bn->nargs == NULL);
-	if ((bn->nargs = strndup(ob->data, ob->size)) == NULL)
-		goto out;
-	rc = 1;
-out:
-	hbuf_free(ob);
-	return rc;
-}
-
 /*
  * Convert a link into a short-link and place the escaped output into a
  * returned string.
@@ -512,25 +541,22 @@ out:
 
 /*
  * Manage hypertext linking with the groff "pdfhref" macro or simply
- * using italics.
- * We use italics because the UR/UE macro doesn't support leading
- * un-spaced content, so "[foo](https://foo.com)" wouldn't work.
- * Until a solution is found, let's just italicise the link text (or
- * link, if no text is found).
- * Return <0 on error (memory), 0 if we've snipped the next node, >0
- * otherwise.
+ * using italics.  XXX: use italics because the UR/UE macro doesn't
+ * support leading un-spaced content, so "x[foo](https://foo.com)y"
+ * wouldn't work.  Until a solution is found, let's just italicise the
+ * link text (or link, if no text is found).  Return FALSE on error
+ * (memory), TRUE on success.
  */
-static ssize_t
+static int
 putlink(struct bnodeq *obq, struct nroff *st, 
-	const struct lowdown_buf *link, struct bnodeq *bq,
-	enum halink_type type, const struct lowdown_node *next)
+	const struct lowdown_buf *link, 
+	const struct lowdown_buf *id, 
+	struct bnodeq *bq, enum halink_type type)
 {
-	struct lowdown_buf		*ob = NULL, *tmp = NULL;
-	const struct lowdown_buf	*nbuf;
-	struct bnode			*bn, *prev;
-	size_t				 sz, i;
-	int				 rc = 0;
-	ssize_t				 ret = 0;
+	struct lowdown_buf	*ob = NULL;
+	struct bnode		*bn;
+	size_t			 i;
+	int			 rc = 0, local = 0;
 
 	/*
 	 * For -Tman or without .pdfhref, format the link as-is, with
@@ -570,7 +596,7 @@ putlink(struct bnodeq *obq, struct nroff *st,
 			rc = 1;
 			goto out;
 		}
-		if (bqueue_span(obq, " (") == NULL)
+		if (bqueue_span(obq, " <") == NULL)
 			goto out;
 		st->fonts[NFONT_ITALIC]++;
 		if (!bqueue_font(st, obq, 0))
@@ -589,7 +615,7 @@ putlink(struct bnodeq *obq, struct nroff *st,
 		st->fonts[NFONT_ITALIC]--;
 		if (!bqueue_font(st, obq, 1))
 			goto out;
-		if (bqueue_span(obq, ")") == NULL)
+		if (bqueue_span(obq, ">") == NULL)
 			goto out;
 		rc = 1;
 		goto out;
@@ -600,73 +626,16 @@ putlink(struct bnodeq *obq, struct nroff *st,
 	if ((ob = hbuf_new(32)) == NULL)
 		goto out;
 
-	/*
-	 * If we're preceded by normal text that doesn't end with space,
-	 * then put that text into the "-P" (prefix) argument.
-	 */
-
-	prev = TAILQ_LAST(obq, bnodeq);
-	if (prev != NULL &&
-	    prev->scope == BSCOPE_SPAN &&
-	    prev->buf != NULL &&
-	    strlen(prev->buf) > 0 && !isspace
-	    ((unsigned char)prev->buf[strlen(prev->buf) - 1])) {
-		sz = strlen(prev->buf);
-		while (sz && !isspace
-		       ((unsigned char)prev->buf[sz - 1]))
-			sz--;
-		assert(sz != strlen(prev->buf));
-
-		if (!HBUF_PUTSL(ob, "-P \""))
-			goto out;
-		if (!hesc_nroff(ob, &prev->buf[sz],
-		    strlen(prev->buf) - sz, 1, 0, 1))
-			goto out;
-		if (!HBUF_PUTSL(ob, "\" "))
-			goto out;
-
-		if ((tmp = hbuf_new(32)) == NULL)
-			goto out;
-		if (!hesc_nroff(tmp, prev->buf, sz, 1, 0, 1))
-			goto out;
-		assert(prev->nbuf == NULL);
-		prev->nbuf = strndup(tmp->data, tmp->size);
-		if (prev->nbuf == NULL)
-			goto out;
-		free(prev->buf);
-		prev->buf = NULL;
-	}
-
-	/* 
-	 * If we're followed by text flush against the node, "chop" the
-	 * node until the next white-space.  This will be handed to the
-	 * next rndr() invocation.
-	 */
-
-	if (next != NULL && 
-	    next->type == LOWDOWN_NORMAL_TEXT) {
-		nbuf = &next->rndr_normal_text.text;
-		for (sz = 0; sz < nbuf->size; sz++)
-			if (isspace((unsigned char)nbuf->data[sz]))
-				break;
-		if (sz > 0) {
-			if (!HBUF_PUTSL(ob, "-A \""))
-				goto out;
-			if (!hesc_nroff(ob, nbuf->data, sz, 1, 0, 1))
-				goto out;
-			if (!HBUF_PUTSL(ob, "\" "))
-				goto out;
-			ret = sz;
-		}
-	}
-
 	/* Encode the URL. */
 
+	local = type != HALINK_EMAIL &&
+		link->size && link->data[0] == '#';
+
 	if (!HBUF_PUTSL(ob, "-D "))
 		goto out;
 	if (type == HALINK_EMAIL && !HBUF_PUTSL(ob, "mailto:"))
 		goto out;
-	for (i = 0; i < link->size; i++) {
+	for (i = local ? 1 : 0; i < link->size; i++) {
 		if (!isprint((unsigned char)link->data[i]) ||
 		    strchr("<>\\^`{|}\"", link->data[i]) != NULL) {
 			if (!hbuf_printf(ob, "%%%.2X", link->data[i]))
@@ -675,34 +644,53 @@ putlink(struct bnodeq *obq, struct nroff *st,
 			goto out;
 	}
 
-	if (!HBUF_PUTSL(ob, " "))
+	if (!HBUF_PUTSL(ob, " -- "))
 		goto out;
 	if (bq == NULL && !hbuf_putb(ob, link))
 		goto out;
 	else if (bq != NULL && !bqueue_flush(ob, bq, 1))
 		goto out;
-	if ((bn = bqueue_block(obq, ".pdfhref W")) == NULL)
+
+	/*
+	 * If we have an ID, emit it before the link part.  This is
+	 * important because this isn't printed, so using "-A \c" will
+	 * have no effect, so that's used on the subsequent link.
+	 */
+
+	if (id != NULL && id->size > 0) {
+		bn = bqueue_semiblock(obq, ".pdfhref M");
+		if (bn == NULL)
+			goto out;
+		bn->args = strndup(id->data, id->size);
+		if (bn->args == NULL)
+			goto out;
+	}
+
+	/* Finally, emit the link contents. */
+
+	bn = local ? 
+		bqueue_semiblock(obq, ".pdfhref L") :
+		bqueue_semiblock(obq, ".pdfhref W");
+	if (bn == NULL)
 		goto out;
 	if ((bn->nargs = strndup(ob->data, ob->size)) == NULL)
 		goto out;
 
 	rc = 1;
 out:
-	hbuf_free(tmp);
 	hbuf_free(ob);
-	return rc ? ret : -1;
+	return rc;
 }
 
 /*
- * Return <0 on failure, 0 to remove next, >0 otherwise.
+ * Return FALSE on failure, TRUE on success.
  */
-static ssize_t
+static int
 rndr_autolink(struct nroff *st, struct bnodeq *obq,
-	const struct rndr_autolink *param, 
-	const struct lowdown_node *next)
+	const struct rndr_autolink *param)
 {
 
-	return putlink(obq, st, &param->link, NULL, param->type, next);
+	return putlink(obq, st, &param->link, NULL, NULL, param->type);
 }
 
 static int
@@ -749,21 +737,36 @@ rndr_blockcode(const struct nroff *st, struct bnodeq *obq,
 static int
 rndr_definition_title(struct bnodeq *obq, struct bnodeq *bq)
 {
-	struct bnode	*bn;
 
-	if ((bn = bqueue_block(obq, ".IP")) == NULL)
+	if (bqueue_block(obq, ".LP") == NULL)
 		return 0;
-	return bqueue_to_nargs(bn, bq, 1);
+	TAILQ_CONCAT(obq, bq, entries);
+	return 1;
 }
 
 static int
 rndr_definition_data(struct bnodeq *obq, struct bnodeq *bq)
 {
-	/* 
-	 * Strip out leading paragraphs.
-	 * XXX: shouldn't these all be handled by the child list item?
+	/*
+	 * The IP creates an empty vertical space til I figure out a
+	 * better way to do hanging lists, so account for it by backing
+	 * up first.
+	 *
+	 * XXX: this produces different results on mandoc and groff as
+	 * of 2022-02-19: mandoc backs up one space, while groff backs
+	 * up two.  The groff behaviour is what we want, so that the
+	 * text is flush on the next line, but this is good enough.
 	 */
 
+	if (bqueue_block(obq, ".if n \\\n.sp -1v") == NULL)
+		return 0;
+	if (bqueue_block(obq, ".if t \\\n.sp -0.25v\n") == NULL)
+		return 0;
+	if (bqueue_block(obq, ".IP \"\" \\*(PI") == NULL)
+		return 0;
+
+	/* Strip out leading paragraphs. */
+
 	bqueue_strip_paras(bq);
 	TAILQ_CONCAT(obq, bq, entries);
 	return 1;
@@ -775,7 +778,7 @@ rndr_list(struct nroff *st, struct bnodeq *obq,
 {
 	/* 
 	 * If we have a nested list, we need to use RS/RE to indent the
-	 * nested component.  Otherwise the `IP` used for the titles and
+	 * nested component.  Otherwise the "IP" used for the titles and
 	 * contained paragraphs won't indent properly.
 	 */
 
@@ -825,24 +828,22 @@ rndr_linebreak(struct bnodeq *obq)
 
 static int
 rndr_header(struct nroff *st, struct bnodeq *obq,
-	struct bnodeq *bq, const struct rndr_header *param)
+	struct bnodeq *bq, const struct lowdown_node *n)
 {
-	ssize_t			 level;
-	struct bnode		*bn;
+	ssize_t				 level;
+	struct bnode			*bn;
+	struct lowdown_buf		*buf = NULL;
+	const struct lowdown_buf	*nbuf;
+	int			 	 rc = 0;
 
-	level = (ssize_t)param->level + st->headers_offs;
+	level = (ssize_t)n->rndr_header.level + st->headers_offs;
 	if (level < 1)
 		level = 1;
 
 	/*
 	 * For man(7), we use SH for the first-level section, SS for
-	 * other sections.
-	 * FIXME: use PP then italics or something for third-level etc.
-	 * For ms(7), just use SH.
-	 * If we're using ms(7) w/groff extensions and w/o numbering,
-	 * used the numbered version of the SH macro.
-	 * If we're numbered ms(7), use NH.
-	 * With groff extensions, use XN (-mspdf).
+	 * other sections.  TODO: use PP then italics or something for
+	 * third-level etc.
 	 */
 
 	if (st->man) {
@@ -851,47 +852,100 @@ rndr_header(struct nroff *st, struct bnodeq *obq,
 			bqueue_block(obq, ".SS");
 		if (bn == NULL)
 			return 0;
-		return bqueue_to_nargs(bn, bq, 0);
+		TAILQ_CONCAT(obq, bq, entries);
+		st->post_para = 1;
+		return 1;
 	} 
 
-	if (st->flags & LOWDOWN_NROFF_NUMBERED)
-		bn = bqueue_block(obq, ".NH");
-	else
-		bn = bqueue_block(obq, ".SH");
+	/*
+	 * If we're using ms(7) w/groff extensions and w/o numbering,
+	 * used the numbered version of the SH macro.
+	 * If we're numbered ms(7), use NH.
+	 */
+
+	bn = (st->flags & LOWDOWN_NROFF_NUMBERED) ?
+		bqueue_block(obq, ".NH") : bqueue_block(obq, ".SH");
 	if (bn == NULL)
-		return 0;
+		goto out;
 
 	if ((st->flags & LOWDOWN_NROFF_NUMBERED) ||
 	    (st->flags & LOWDOWN_NROFF_GROFF)) 
 		if (asprintf(&bn->nargs, "%zd", level) == -1) {
 			bn->nargs = NULL;
-			return 0;
+			goto out;
 		}
 
-	/* Used in -mspdf output for creating a TOC. */
+	TAILQ_CONCAT(obq, bq, entries);
+	st->post_para = 1;
+
+	/*
+	 * Used in -mspdf output for creating a TOC and intra-document
+	 * linking.
+	 */
 
 	if (st->flags & LOWDOWN_NROFF_GROFF) {
-		if ((bn = bqueue_block(obq, ".XN")) == NULL)
-			return 0;
-		if (!bqueue_to_nargs(bn, bq, 0))
-			return 0;
-	} else
-		TAILQ_CONCAT(obq, bq, entries);
+		if ((buf = hbuf_new(32)) == NULL)
+			goto out;
+		if (!hbuf_extract_text(buf, n))
+			goto out;
 
-	st->post_para = 1;
-	return 1;
+		if ((bn = bqueue_block(obq, ".pdfhref")) == NULL)
+			goto out;
+		if (asprintf(&bn->nargs, "O %zd", level) == -1) {
+			bn->nargs = NULL;
+			goto out;
+		}
+
+		/*
+		 * No need to quote: quotes will be converted by
+		 * escaping into roff.
+		 */
+
+		bn->args = strndup(buf->data, buf->size);
+		if (bn->args == NULL)
+			goto out;
+
+		if ((bn = bqueue_block(obq, ".pdfhref M")) == NULL)
+			goto out;
+
+		/*
+		 * If the identifier comes from the user, we need to
+		 * escape it accordingly; otherwise, use it directly as
+		 * the hbuf_id() function will take care of it.
+		 */
+
+		if (n->rndr_header.attr_id.size) {
+			bn->args = strndup
+				(n->rndr_header.attr_id.data,
+				 n->rndr_header.attr_id.size);
+			if (bn->args == NULL)
+				goto out;
+		} else {
+			nbuf = hbuf_id(buf, NULL, &st->headers_used);
+			if (nbuf == NULL)
+				goto out;
+			bn->nargs = strndup(nbuf->data, nbuf->size);
+			if (bn->nargs == NULL)
+				goto out;
+		}
+	}
+
+	rc = 1;
+out:
+	hbuf_free(buf);
+	return rc;
 }
 
 /*
- * Return <0 on failure, 0 to remove next, >0 otherwise.
+ * Return FALSE on failure, TRUE on success.
  */
 static ssize_t
 rndr_link(struct nroff *st, struct bnodeq *obq, struct bnodeq *bq,
-	const struct rndr_link *param, 
-	const struct lowdown_node *next)
+	const struct rndr_link *param)
 {
 
-	return putlink(obq, st, &param->link, bq, HALINK_NORMAL, next);
+	return putlink(obq, st, &param->link,
+		&param->attr_id, bq, HALINK_NORMAL);
 }
 
 static int
@@ -930,10 +984,10 @@ rndr_listitem(struct bnodeq *obq, const struct lowdown_node *n,
 	 * list item that comes after us (i.e., anything after us).
 	 */
 
-	if (n->rndr_listitem.flags & HLIST_FL_BLOCK)
+	if ((n->rndr_listitem.flags & HLIST_FL_BLOCK) ||
+	    (n->rndr_listitem.flags & HLIST_FL_DEF))
 		return 1;
-	if (n->rndr_listitem.flags & HLIST_FL_DEF)
-		n = n->parent;
+
 	if (TAILQ_NEXT(n, entries) != NULL) {
 		if (bqueue_block(obq, ".if n \\\n.sp -1") == NULL)
 			return 0;
@@ -948,30 +1002,25 @@ static int
 rndr_paragraph(struct nroff *st, const struct lowdown_node *n,
 	struct bnodeq *obq, struct bnodeq *nbq)
 {
-	const struct lowdown_node	*prev;
-	struct bnode			*bn;
+	struct bnode	*bn;
 
 	/* 
-	 * We don't just blindly print a paragraph macro: it depends
-	 * upon what came before.  If we're following a HEADER, don't do
-	 * anything at all.  If we're in a list item, make sure that we
-	 * don't reset our text indent by using an `IP`.
+	 * Subsequent paragraphs get a PP for the indentation; otherwise, use
+	 * LP and forego the indentation.  If we're in a list item, make sure
+	 * that we don't reset our text indent by using an "IP".
 	 */
 
-	prev = TAILQ_PREV(n, lowdown_nodeq, entries);
-	if (!st->man || prev == NULL || prev->type != LOWDOWN_HEADER) {
-		for ( ; n != NULL; n = n->parent)
-			if (n->type == LOWDOWN_LISTITEM)
-				break;
-		if (n != NULL)
-			bn = bqueue_block(obq, ".IP");
-		else if (st->post_para)
-			bn = bqueue_block(obq, ".LP");
-		else
-			bn = bqueue_block(obq, ".PP");
-		if (bn == NULL)
-			return 0;
-	}
+	for ( ; n != NULL; n = n->parent)
+		if (n->type == LOWDOWN_LISTITEM)
+			break;
+	if (n != NULL)
+		bn = bqueue_block(obq, ".IP");
+	else if (st->post_para)
+		bn = bqueue_block(obq, ".LP");
+	else
+		bn = bqueue_block(obq, ".PP");
+	if (bn == NULL)
+		return 0;
 
 	TAILQ_CONCAT(obq, nbq, entries);
 	st->post_para = 0;
@@ -995,19 +1044,30 @@ rndr_raw_block(const struct nroff *st,
 }
 
 static int
-rndr_hrule(const struct nroff *st, struct bnodeq *obq)
+rndr_hrule(struct nroff *st, struct bnodeq *obq)
 {
-	/*
-	 * I'm not sure how else to do horizontal lines.
-	 * The LP is to reset the margins.
-	 */
+
+	/* The LP is to reset the margins. */
 
 	if (bqueue_block(obq, ".LP") == NULL)
 		return 0;
-	if (!st->man && 
-	    bqueue_block(obq, "\\l\'\\n(.lu-\\n(\\n[.in]u\'") == NULL)
-		return 0;
-	return 1;
+
+	/* Set post_para so we get a following LP not PP. */
+
+	st->post_para = 1;
+
+	if (st->man)
+		return bqueue_block(obq, "\\l\'2i'") != NULL;
+
+	return bqueue_block(obq,
+	    ".ie d HR \\{\\\n"
+	    ".HR\n"
+	    "\\}\n"
+	    ".el \\{\\\n"
+	    ".sp 1v\n"
+	    "\\l'\\n(.lu'\n"
+	    ".sp 1v\n"
+	    ".\\}") != NULL;
 }
 
 static int
@@ -1018,63 +1078,58 @@ rndr_image(struct nroff *st, struct bnodeq *obq,
 	size_t		 sz;
 	struct bnode	*bn;
 
-	/* In -Tman, we have no images: treat as a link. */
-
-	if (st->man) {
-		st->fonts[NFONT_BOLD]++;
-		if (!bqueue_font(st, obq, 0))
-			return 0;
-		if ((bn = bqueue_span(obq, NULL)) == NULL)
-			return 0;
-		bn->buf = strndup(param->alt.data, param->alt.size);
-		if (bn->buf == NULL)
-			return 0;
-		st->fonts[NFONT_BOLD]--;
-		if (!bqueue_font(st, obq, 1))
-			return 0;
-		if (st->flags & LOWDOWN_NROFF_NOLINK)
-			return bqueue_span(obq, " (Image)") != NULL;
-		if (bqueue_span(obq, " (Image: ") == NULL)
-			return 0;
-		st->fonts[NFONT_ITALIC]++;
-		if (!bqueue_font(st, obq, 0))
-			return 0;
-		if ((bn = bqueue_span(obq, NULL)) == NULL)
-			return 0;
-		if (st->flags & LOWDOWN_NROFF_SHORTLINK) {
-			bn->nbuf = hbuf2shortlink(&param->link);
-			if (bn->nbuf == NULL)
-				return 0;
-		} else {
-			bn->buf = strndup(param->link.data, param->link.size);
-			if (bn->buf == NULL)
-				return 0;
+	if (!st->man) {
+		cp = memrchr(param->link.data, '.', param->link.size);
+		if (cp != NULL) {
+			cp++;
+			sz = param->link.size - (cp - param->link.data);
+			if ((sz == 2 && memcmp(cp, "ps", 2) == 0) ||
+			    (sz == 3 && memcmp(cp, "eps", 3) == 0)) {
+				bn = bqueue_block(obq, ".PSPIC");
+				if (bn == NULL)
+					return 0;
+				bn->args = strndup(param->link.data,
+					param->link.size);
+				return bn->args != NULL;
+			}
 		}
-		st->fonts[NFONT_ITALIC]--;
-		if (!bqueue_font(st, obq, 1))
-			return 0;
-		return bqueue_span(obq, ")") != NULL;
 	}
 
-	/* Are we suffixed with ps or eps? */
+	/* In -Tman, we have no images: treat as a link. */
 
-	cp = memrchr(param->link.data, '.', param->link.size);
-	if (cp == NULL)
-		return 1;
-	cp++;
-	sz = param->link.size - (cp - param->link.data);
-	if (sz == 0)
-		return 1;
-	if (!(sz == 2 && memcmp(cp, "ps", 2) == 0) &&
-	    !(sz == 3 && memcmp(cp, "eps", 3) == 0))
-		return 1;
-
-	/* If so, use a PSPIC. */
-
-	if ((bn = bqueue_block(obq, ".PSPIC")) == NULL)
+	st->fonts[NFONT_BOLD]++;
+	if (!bqueue_font(st, obq, 0))
 		return 0;
-	bn->args = strndup(param->link.data, param->link.size);
-	return bn->args != NULL;
+	if ((bn = bqueue_span(obq, NULL)) == NULL)
+		return 0;
+	bn->buf = strndup(param->alt.data, param->alt.size);
+	if (bn->buf == NULL)
+		return 0;
+	st->fonts[NFONT_BOLD]--;
+	if (!bqueue_font(st, obq, 1))
+		return 0;
+	if (st->flags & LOWDOWN_NROFF_NOLINK)
+		return bqueue_span(obq, " (Image)") != NULL;
+	if (bqueue_span(obq, " (Image: ") == NULL)
+		return 0;
+	st->fonts[NFONT_ITALIC]++;
+	if (!bqueue_font(st, obq, 0))
+		return 0;
+	if ((bn = bqueue_span(obq, NULL)) == NULL)
+		return 0;
+	if (st->flags & LOWDOWN_NROFF_SHORTLINK) {
+		bn->nbuf = hbuf2shortlink(&param->link);
+		if (bn->nbuf == NULL)
+			return 0;
+	} else {
+		bn->buf = strndup(param->link.data, param->link.size);
+		if (bn->buf == NULL)
+			return 0;
+	}
+	st->fonts[NFONT_ITALIC]--;
+	if (!bqueue_font(st, obq, 1))
+		return 0;
+	return bqueue_span(obq, ")") != NULL;
 }
 
 static int
@@ -1224,30 +1279,9 @@ rndr_superscript(struct bnodeq *obq, struct bnodeq *bq)
 	return bqueue_span(obq, "\\s+3\\d") != NULL;
 }
 
-static int
-rndr_footnotes(const struct nroff *st,
-	struct bnodeq *obq, struct bnodeq *bq)
-{
-
-	/* Put a horizontal line in the case of man(7). */
-
-	if (st->man) {
-		if (bqueue_block(obq, ".LP") == NULL)
-			return 0;
-		if (bqueue_block(obq, ".sp 3") == NULL)
-			return 0;
-		if (bqueue_block(obq, "\\l\'2i'") == NULL)
-			return 0;
-	}
-
-	TAILQ_CONCAT(obq, bq, entries);
-	return 1;
-}
-
 static int
 rndr_footnote_def(const struct nroff *st, struct bnodeq *obq,
-	struct bnodeq *bq, const struct lowdown_node *n,
-	const struct rndr_footnote_def *param)
+	struct bnodeq *bq, size_t num)
 {
 	struct bnode	*bn;
 
@@ -1261,16 +1295,8 @@ rndr_footnote_def(const struct nroff *st, struct bnodeq *obq,
 	if (!st->man) {
 		if (bqueue_block(obq, ".FS") == NULL)
 			return 0;
-		if ((n->chng == LOWDOWN_CHNG_INSERT ||
-		     n->chng == LOWDOWN_CHNG_DELETE) &&
-		    !bqueue_colour(obq, n->chng, 0))
-			return 0;
 		bqueue_strip_paras(bq);
 		TAILQ_CONCAT(obq, bq, entries);
-		if ((n->chng == LOWDOWN_CHNG_INSERT ||
-		     n->chng == LOWDOWN_CHNG_DELETE) &&
-		    !bqueue_colour(obq, n->chng, 1))
-			return 0;
 		return bqueue_block(obq, ".FE") != NULL;
 	}
 
@@ -1281,33 +1307,49 @@ rndr_footnote_def(const struct nroff *st, struct bnodeq *obq,
 
 	if (bqueue_block(obq, ".LP") == NULL)
 		return 0;
-	if ((n->chng == LOWDOWN_CHNG_INSERT ||
-	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    !bqueue_colour(obq, n->chng, 0))
-		return 0;
-
 	if ((bn = bqueue_span(obq, NULL)) == NULL)
 		return 0;
 	if (asprintf(&bn->nbuf, 
-	    "\\0\\fI\\u\\s-3%zu\\s+3\\d\\fP\\0", param->num) == -1) {
+	    "\\0\\fI\\u\\s-3%zu\\s+3\\d\\fP\\0", num) == -1) {
 		bn->nbuf = NULL;
 		return 0;
 	}
-
 	bqueue_strip_paras(bq);
 	TAILQ_CONCAT(obq, bq, entries);
-	if ((n->chng == LOWDOWN_CHNG_INSERT ||
-	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    !bqueue_colour(obq, n->chng, 1))
-		return 0;
 	return 1;
 }
 
 static int
-rndr_footnote_ref(const struct nroff *st,
-	struct bnodeq *obq, const struct rndr_footnote_ref *param)
+rndr_footnotes(const struct nroff *st, struct bnodeq *obq)
+{
+	size_t	 i;
+
+	if (st->footsz == 0)
+		return 1;
+
+	if (st->man) {
+		if (bqueue_block(obq, ".LP") == NULL)
+			return 0;
+		if (bqueue_block(obq, ".sp 3") == NULL)
+			return 0;
+		if (bqueue_block(obq, "\\l\'2i'") == NULL)
+			return 0;
+	}
+
+	for (i = 0; i < st->footsz; i++)
+		if (!rndr_footnote_def(st, obq, st->foots[i], i + 1))
+			return 0;
+
+	return 1;
+}
+
+static int
+rndr_footnote_ref(struct nroff *st, struct bnodeq *obq,
+	struct bnodeq *bq)
 {
 	struct bnode	*bn;
+	void		*pp;
+	size_t		 num = st->footsz;
 
 	/* 
 	 * Use groff_ms(7)-style automatic footnoting, else just put a
@@ -1317,33 +1359,79 @@ rndr_footnote_ref(const struct nroff *st,
 	if ((bn = bqueue_span(obq, NULL)) == NULL)
 		return 0;
 
-	if (st->man) {
-		if (asprintf(&bn->nbuf, 
-		    "\\u\\s-3%zu\\s+3\\d", param->num) == -1)
-			bn->nbuf = NULL;
-	} else
+	if (!st->man)
 		bn->nbuf = strdup("\\**");
+	else if (asprintf(&bn->nbuf, 
+		 "\\u\\s-3%zu\\s+3\\d", num + 1) == -1)
+		bn->nbuf = NULL;
 
-	return bn->nbuf != NULL;
+	if (bn->nbuf == NULL)
+		return 0;
+
+	/*
+	 * For -Tman, queue the footnote for printing at the end of the
+	 * document.  For -Tms, emit it now in a FS/FE block.
+	 */
+
+	if (st->man) {
+		pp = recallocarray(st->foots, st->footsz,
+			st->footsz + 1, sizeof(struct bnodeq *));
+		if (pp == NULL)
+			return 0;
+		st->foots = pp;
+		st->foots[st->footsz++] = malloc(sizeof(struct bnodeq));
+		if (st->foots[num] == NULL)
+			return 0;
+		TAILQ_INIT(st->foots[num]);
+		TAILQ_CONCAT(st->foots[num], bq, entries);
+		return 1;
+	} else {
+		if (bqueue_block(obq, ".FS") == NULL)
+			return 0;
+		bqueue_strip_paras(bq);
+		TAILQ_CONCAT(obq, bq, entries);
+		return bqueue_block(obq, ".FE") != NULL;
+	}
 }
 
 static int
 rndr_entity(const struct nroff *st,
 	struct bnodeq *obq, const struct rndr_entity *param)
 {
-	int32_t		 ent;
-	struct bnode	*bn;
 	char		 buf[32];
+	const char	*ent;
+	struct bnode	*bn;
+	int32_t		 iso;
+	size_t		 sz;
 
-	if ((ent = entity_find_iso(&param->text)) > 0) {
+	/*
+	 * Handle named entities if "ent" is non-NULL, use unicode
+	 * escapes for values above 126, and just the regular character
+	 * if within the ASCII set.
+	 */
+
+	if ((ent = entity_find_nroff(&param->text, &iso)) != NULL) {
+		sz = strlen(ent);
+		if (sz == 1)
+			snprintf(buf, sizeof(buf), "\\%s", ent);
+		else if (sz == 2)
+			snprintf(buf, sizeof(buf), "\\(%s", ent);
+		else
+			snprintf(buf, sizeof(buf), "\\[%s]", ent);
+		return bqueue_span(obq, buf) != NULL;
+	} else if (iso > 0 && iso > 126) {
 		if (st->flags & LOWDOWN_NROFF_GROFF)
 			snprintf(buf, sizeof(buf), "\\[u%.4llX]", 
-				(unsigned long long)ent);
+				(unsigned long long)iso);
 		else
 			snprintf(buf, sizeof(buf), "\\U\'%.4llX\'", 
-				(unsigned long long)ent);
+				(unsigned long long)iso);
 		return bqueue_span(obq, buf) != NULL;
-	} 
+	} else if (iso > 0) {
+		snprintf(buf, sizeof(buf), "%c", iso);
+		return bqueue_span(obq, buf) != NULL;
+	}
+
 	if ((bn = bqueue_span(obq, NULL)) == NULL)
 		return 0;
 	bn->buf = strndup(param->text.data, param->text.size);
@@ -1352,9 +1440,7 @@ rndr_entity(const struct nroff *st,
 
 /*
  * Split "b" at sequential white-space, outputting the results in the
- * line-based "env" macro.
- * The content in "b" has already been escaped, so there's no need to do
- * anything but manage white-space.
+ * line-based "env" macro.  The content in "b" has already been escaped.
  */
 static int
 rndr_meta_multi(struct bnodeq *obq, const char *b, const char *env)
@@ -1391,13 +1477,17 @@ rndr_meta_multi(struct bnodeq *obq, const char *b, const char *env)
 			return 0;
 		if ((bn = bqueue_span(obq, NULL)) == NULL)
 			return 0;
-		if ((bn->buf = strndup(start, sz)) == NULL)
+		if ((bn->nbuf = strndup(start, sz)) == NULL)
 			return 0;
 	}
 
 	return 1;
 }
 
+/*
+ * Fill "mq" by serialising child nodes into strings.  The serialised
+ * strings are escaped.
+ */
 static int
 rndr_meta(struct nroff *st, const struct bnodeq *bq, 
 	struct lowdown_metaq *mq, const struct rndr_meta *params)
@@ -1417,7 +1507,7 @@ rndr_meta(struct nroff *st, const struct bnodeq *bq,
 
 	if ((ob = hbuf_new(32)) == NULL)
 		return 0;
-	if (!bqueue_flush(ob, bq, 0)) {
+	if (!bqueue_flush(ob, bq, 1)) {
 		hbuf_free(ob);
 		return 0;
 	}
@@ -1502,7 +1592,7 @@ rndr_doc_header(const struct nroff *st,
 				".ds LF Copyright \\(co");
 			if (bn == NULL)
 				goto out;
-			if ((bn->args = strdup(copy)) == NULL)
+			if ((bn->nargs = strdup(copy)) == NULL)
 				goto out;
 		}
 		if (date != NULL) {
@@ -1512,14 +1602,15 @@ rndr_doc_header(const struct nroff *st,
 				bn = bqueue_block(obq, ".DA");
 			if (bn == NULL)
 				goto out;
-			if ((bn->args = strdup(date)) == NULL)
+			if ((bn->nargs = strdup(date)) == NULL)
 				goto out;
 		}
+
 		if (bqueue_block(obq, ".TL") == NULL)
 			goto out;
 		if ((bn = bqueue_span(obq, NULL)) == NULL)
 			goto out;
-		if ((bn->buf = strdup(title)) == NULL)
+		if ((bn->nbuf = strdup(title)) == NULL)
 			goto out;
 		if (!rndr_meta_multi(obq, author, "AU"))
 			goto out;
@@ -1536,10 +1627,11 @@ rndr_doc_header(const struct nroff *st,
 
 		if ((bn = bqueue_block(obq, ".TH")) == NULL)
 			goto out;
+
 		if (!hbuf_putc(ob, '"') ||
-		    !hesc_nroff(ob, title, strlen(title), 1, 0, 1) ||
+		    !hesc_nroff(ob, title, strlen(title), 1, 0, 0) ||
 		    !HBUF_PUTSL(ob, "\" \"") ||
-		    !hesc_nroff(ob, sec, strlen(sec), 1, 0, 1) ||
+		    !hesc_nroff(ob, sec, strlen(sec), 1, 0, 0) ||
 		    !hbuf_putc(ob, '"'))
 			goto out;
 
@@ -1551,7 +1643,7 @@ rndr_doc_header(const struct nroff *st,
 		if (!HBUF_PUTSL(ob, " \""))
 			goto out;
 		if (date != NULL &&
-		    !hesc_nroff(ob, date, strlen(date), 1, 0, 1))
+		    !hesc_nroff(ob, date, strlen(date), 1, 0, 0))
 			goto out;
 		if (!HBUF_PUTSL(ob, "\""))
 			goto out;
@@ -1566,14 +1658,14 @@ rndr_doc_header(const struct nroff *st,
 			if (!HBUF_PUTSL(ob, " \""))
 				goto out;
 			if (source != NULL && !hesc_nroff
-			    (ob, source, strlen(source), 1, 0, 1))
+			    (ob, source, strlen(source), 1, 0, 0))
 				goto out;
 			if (!HBUF_PUTSL(ob, "\""))
 				goto out;
 			if (!HBUF_PUTSL(ob, " \""))
 				goto out;
 			if (volume != NULL && !hesc_nroff
-			    (ob, volume, strlen(volume), 1, 0, 1))
+			    (ob, volume, strlen(volume), 1, 0, 0))
 				goto out;
 			if (!HBUF_PUTSL(ob, "\""))
 				goto out;
@@ -1594,13 +1686,12 @@ out:
  * Return what (if anything) we should chop from the next node or <0 on
  * failure.
  */
-static ssize_t
+static int
 rndr(struct lowdown_metaq *mq, struct nroff *st,
-	const struct lowdown_node *n, struct bnodeq *obq, size_t chop)
+	const struct lowdown_node *n, struct bnodeq *obq)
 {
 	const struct lowdown_node	*child;
 	int				 rc = 1;
-	ssize_t				 keepnext, ret = -1;
 	enum nfont			 fonts[NFONT__MAX];
 	struct bnodeq			 tmpbq;
 	struct bnode			*bn;
@@ -1609,7 +1700,6 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 
 	if ((n->chng == LOWDOWN_CHNG_INSERT ||
 	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !bqueue_colour(obq, n->chng, 0))
 		goto out;
 
@@ -1648,14 +1738,10 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 		break;
 	}
 
-	keepnext = 0;
-	TAILQ_FOREACH(child, &n->children, entries) {
-		keepnext = rndr(mq, st, child, &tmpbq, keepnext);
-		if (keepnext < 0)
+	TAILQ_FOREACH(child, &n->children, entries)
+		if (!rndr(mq, st, child, &tmpbq))
 			goto out;
-	}
 
-	ret = 0;
 	switch (n->type) {
 	case LOWDOWN_BLOCKCODE:
 		rc = rndr_blockcode(st, obq, &n->rndr_blockcode);
@@ -1680,7 +1766,7 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 			rc = rndr_meta(st, &tmpbq, mq, &n->rndr_meta);
 		break;
 	case LOWDOWN_HEADER:
-		rc = rndr_header(st, obq, &tmpbq, &n->rndr_header);
+		rc = rndr_header(st, obq, &tmpbq, n);
 		break;
 	case LOWDOWN_HRULE:
 		rc = rndr_hrule(st, obq);
@@ -1707,19 +1793,15 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 	case LOWDOWN_TABLE_CELL:
 		rc = rndr_table_cell(obq, &tmpbq, &n->rndr_table_cell);
 		break;
-	case LOWDOWN_FOOTNOTES_BLOCK:
-		rc = rndr_footnotes(st, obq, &tmpbq);
-		break;
-	case LOWDOWN_FOOTNOTE_DEF:
-		rc = rndr_footnote_def(st, obq, 
-			&tmpbq, n, &n->rndr_footnote_def);
+	case LOWDOWN_ROOT:
+		TAILQ_CONCAT(obq, &tmpbq, entries);
+		rc = rndr_footnotes(st, obq);
 		break;
 	case LOWDOWN_BLOCKHTML:
 		rc = rndr_raw_block(st, obq, &n->rndr_blockhtml);
 		break;
 	case LOWDOWN_LINK_AUTO:
-		ret = rndr_autolink(st, obq, 
-			&n->rndr_autolink, TAILQ_NEXT(n, entries));
+		rc = rndr_autolink(st, obq, &n->rndr_autolink);
 		break;
 	case LOWDOWN_CODESPAN:
 		rc = rndr_codespan(obq, &n->rndr_codespan);
@@ -1731,21 +1813,18 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 		rc = rndr_linebreak(obq);
 		break;
 	case LOWDOWN_LINK:
-		ret = rndr_link(st, obq, &tmpbq, 
-			&n->rndr_link, TAILQ_NEXT(n, entries));
+		rc = rndr_link(st, obq, &tmpbq, &n->rndr_link);
 		break;
 	case LOWDOWN_SUPERSCRIPT:
 		rc = rndr_superscript(obq, &tmpbq);
 		break;
-	case LOWDOWN_FOOTNOTE_REF:
-		rc = rndr_footnote_ref(st, obq, &n->rndr_footnote_ref);
+	case LOWDOWN_FOOTNOTE:
+		rc = rndr_footnote_ref(st, obq, &tmpbq);
 		break;
 	case LOWDOWN_RAW_HTML:
 		rc = rndr_raw_html(st, obq, &n->rndr_raw_html);
 		break;
 	case LOWDOWN_NORMAL_TEXT:
-		if (chop == n->rndr_normal_text.text.size)
-			break;
 		if ((bn = bqueue_span(obq, NULL)) == NULL)
 			goto out;
 		bn->buf = strndup
@@ -1753,7 +1832,6 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 			 n->rndr_normal_text.text.size);
 		if (bn->buf == NULL)
 			goto out;
-		bn->bufchop = chop;
 		break;
 	case LOWDOWN_ENTITY:
 		rc = rndr_entity(st, obq, &n->rndr_entity);
@@ -1764,8 +1842,6 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 	}
 
 	if (!rc)
-		ret = -1;
-	if (ret < 0)
 		goto out;
 
 	/* Restore the font stack. */
@@ -1778,7 +1854,7 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 	case LOWDOWN_TRIPLE_EMPHASIS:
 		memcpy(st->fonts, fonts, sizeof(fonts));
 		if (!bqueue_font(st, obq, 1)) {
-			ret = -1;
+			rc = 0;
 			goto out;
 		}
 		break;
@@ -1788,15 +1864,15 @@ rndr(struct lowdown_metaq *mq, struct nroff *st,
 
 	if ((n->chng == LOWDOWN_CHNG_INSERT ||
 	     n->chng == LOWDOWN_CHNG_DELETE) &&
-	    n->type != LOWDOWN_FOOTNOTE_DEF &&
 	    !bqueue_colour(obq, n->chng, 1)) {
-		ret = -1;
+		rc = 0;
 		goto out;
 	}
 
+	rc = 1;
 out:
 	bqueue_free(&tmpbq);
-	return ret;
+	return rc;
 }
 
 int
@@ -1807,15 +1883,17 @@ lowdown_nroff_rndr(struct lowdown_buf *ob,
 	struct lowdown_metaq	 metaq;
 	int			 rc = 0;
 	struct bnodeq		 bq;
+	size_t			 i;
 
 	TAILQ_INIT(&metaq);
 	TAILQ_INIT(&bq);
+	TAILQ_INIT(&st->headers_used);
 
 	memset(st->fonts, 0, sizeof(st->fonts));
 	st->headers_offs = 1;
 	st->post_para = 0;
 
-	if (rndr(&metaq, st, n, &bq, 0) >= 0) {
+	if (rndr(&metaq, st, n, &bq)) {
 		if (!bqueue_flush(ob, &bq, 1))
 			goto out;
 		if (ob->size && ob->data[ob->size - 1] != '\n' &&
@@ -1823,9 +1901,19 @@ lowdown_nroff_rndr(struct lowdown_buf *ob,
 			goto out;
 		rc = 1;
 	}
+
 out:
+	for (i = 0; i < st->footsz; i++) {
+		bqueue_free(st->foots[i]);
+		free(st->foots[i]);
+	}
+
+	free(st->foots);
+	st->footsz = 0;
+	st->foots = NULL;
 	lowdown_metaq_free(&metaq);
 	bqueue_free(&bq);
+	hentryq_clear(&st->headers_used);
 	return rc;
 }
 
