@@ -33,23 +33,26 @@
 #include "extern.h"
 
 struct tstack {
-	const struct lowdown_node 	*n; /* node in question */
-	size_t				 lines; /* times emitted */
+	const struct lowdown_node *n; /* node in question */
+	size_t			   lines; /* times emitted */
 };
 
 struct term {
-	unsigned int		 opts; /* oflags from lowdown_cfg */
-	size_t			 col; /* output column from zero */
-	ssize_t			 last_blank; /* line breaks or -1 (start) */
-	struct tstack		*stack; /* stack of nodes */
-	size_t			 stackmax; /* size of stack */
-	size_t			 stackpos; /* position in stack */
-	size_t			 maxcol; /* soft limit */
-	size_t			 hmargin; /* left of content */
-	size_t			 vmargin; /* before/after content */
-	struct lowdown_buf	*tmp; /* for temporary allocations */
-	wchar_t			*buf; /* buffer for counting wchar */
-	size_t			 bufsz; /* size of buf */
+	unsigned int		  opts; /* oflags from lowdown_cfg */
+	size_t			  col; /* output column from zero */
+	ssize_t			  last_blank; /* line breaks or -1 (start) */
+	struct tstack		 *stack; /* stack of nodes */
+	size_t			  stackmax; /* size of stack */
+	size_t			  stackpos; /* position in stack */
+	size_t			  maxcol; /* soft limit */
+	size_t			  hmargin; /* left of content */
+	size_t			  vmargin; /* before/after content */
+	struct lowdown_buf	 *tmp; /* for temporary allocations */
+	wchar_t			 *buf; /* buffer for counting wchar */
+	size_t			  bufsz; /* size of buf */
+	struct lowdown_buf	**foots; /* footnotes */
+	size_t			  footsz; /* footnotes size  */
+	int			  footoff; /* don't collect (tables) */
 };
 
 /*
@@ -95,8 +98,6 @@ static const struct sty *stys[LOWDOWN__MAX] = {
 	NULL, /* LOWDOWN_TABLE_BODY */
 	NULL, /* LOWDOWN_TABLE_ROW */
 	NULL, /* LOWDOWN_TABLE_CELL */
-	NULL, /* LOWDOWN_FOOTNOTES_BLOCK */
-	NULL, /* LOWDOWN_FOOTNOTE_DEF */
 	&sty_blockhtml, /* LOWDOWN_BLOCKHTML */
 	&sty_autolink, /* LOWDOWN_LINK_AUTO */
 	&sty_codespan, /* LOWDOWN_CODESPAN */
@@ -109,14 +110,13 @@ static const struct sty *stys[LOWDOWN__MAX] = {
 	&sty_t_emph, /* LOWDOWN_TRIPLE_EMPHASIS */
 	&sty_strike, /* LOWDOWN_STRIKETHROUGH */
 	NULL, /* LOWDOWN_SUPERSCRIPT */
-	&sty_fref, /* LOWDOWN_FOOTNOTE_REF */
+	NULL, /* LOWDOWN_FOOTNOTE */
 	NULL, /* LOWDOWN_MATH_BLOCK */
 	&sty_rawhtml, /* LOWDOWN_RAW_HTML */
 	NULL, /* LOWDOWN_ENTITY */
 	NULL, /* LOWDOWN_NORMAL_TEXT */
 	NULL, /* LOWDOWN_DOC_HEADER */
 	NULL, /* LOWDOWN_META */
-	NULL /* LOWDOWN_DOC_FOOTER */
 };
 
 /*
@@ -130,8 +130,7 @@ static const struct sty *stys[LOWDOWN__MAX] = {
 /* Forward declaration. */
 
 static int
-rndr(struct lowdown_buf *, struct lowdown_metaq *,
-	struct term *, const struct lowdown_node *);
+rndr(struct lowdown_buf *, struct term *, const struct lowdown_node *);
 
 /*
  * Get the column width of a multi-byte sequence.  The sequence should
@@ -212,6 +211,20 @@ rndr_escape(struct term *term, struct lowdown_buf *out,
 	return cols;
 }
 
+static void
+rndr_free_footnotes(struct term *st)
+{
+	size_t	 i;
+
+	for (i = 0; i < st->footsz; i++)
+		hbuf_free(st->foots[i]);
+
+	free(st->foots);
+	st->foots = NULL;
+	st->footsz = 0;
+	st->footoff = 0;
+}
+
 /*
  * If there's an active style in "s" or s is NULL), then emit an
  * unstyling escape sequence.  Return zero on failure (memory), non-zero
@@ -271,8 +284,8 @@ rndr_buf_style(const struct term *term,
 			return 0;
 	}
 	if (s->bcolour && !(term->opts & LOWDOWN_TERM_NOCOLOUR) &&
-	    ((s->bcolour >= 30 && s->bcolour <= 37) ||
-	     (s->bcolour >= 90 && s->bcolour <= 97))) {
+	    ((s->bcolour >= 40 && s->bcolour <= 47) ||
+	     (s->bcolour >= 100 && s->bcolour <= 107))) {
 		if (has++ && !HBUF_PUTSL(out, ";"))
 			return 0;
 		if (!hbuf_printf(out, "%zu", s->bcolour))
@@ -486,16 +499,23 @@ rndr_buf_startline_prefixes(struct term *term,
 	/*
 	 * Look up the current node in the list of node's we're
 	 * servicing so we can get how many times we've output the
-	 * prefix.
-	 * This is used for (e.g.) lists, where we only output the list
-	 * prefix once.
-	 * FIXME: read backwards for faster perf.
+	 * prefix.  This is used for (e.g.) lists, where we only output
+	 * the list prefix once.  XXX: read backwards for faster perf?
 	 */
 
 	for (i = 0; i <= term->stackpos; i++)
 		if (term->stack[i].n == n)
 			break;
-	assert(i <= term->stackpos);
+
+	/*
+	 * If we can't find the node, then we're in a "faked" context
+	 * like footnotes within a table.  Ignore this.  XXX: is there a
+	 * non-hacky way for this?
+	 */
+
+	if (i > term->stackpos)
+		return 1;
+
 	emit = term->stack[i].lines++;
 
 	/*
@@ -561,16 +581,16 @@ rndr_buf_startline_prefixes(struct term *term,
 			rndr_buf_advance(term, pfx_dli_n.cols);
 		}
 		break;
-	case LOWDOWN_FOOTNOTE_DEF:
+	case LOWDOWN_FOOTNOTE:
 		rndr_node_style_apply(&sinner, &sty_fdef_pfx);
 		if (!rndr_buf_style(term, out, &sinner))
 			return 0;
 		pstyle = 1;
 		if (emit == 0) {
 			if (!hbuf_printf(out, "%2zu. ",
-			     n->rndr_footnote_def.num))
+			    term->footsz + 1))
 				return 0;
-			len = rndr_numlen(n->rndr_footnote_def.num);
+			len = rndr_numlen(term->footsz + 1);
 			if (len + 2 > pfx_fdef_1.cols)
 				len += 2;
 			else
@@ -951,14 +971,13 @@ rndr_stackpos_init(struct term *p, const struct lowdown_node *n)
  * Return zero on failure (memory), non-zero on success.
  */
 static int
-rndr_table(struct lowdown_buf *ob, struct lowdown_metaq *mq,
-	struct term *p, const struct lowdown_node *n)
+rndr_table(struct lowdown_buf *ob, struct term *p,
+	const struct lowdown_node *n)
 {
 	size_t				*widths = NULL;
 	const struct lowdown_node	*row, *top, *cell;
-	struct lowdown_buf		*celltmp = NULL,
-					*rowtmp = NULL;
-	size_t				 col, i, j, maxcol, sz;
+	struct lowdown_buf		*celltmp = NULL, *rowtmp = NULL;
+	size_t				 col, i, j, maxcol, sz, footsz;
 	ssize_t			 	 last_blank;
 	unsigned int			 flags;
 	int				 rc = 0;
@@ -975,9 +994,15 @@ rndr_table(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 
 	/*
 	 * Begin by counting the number of printable columns in each
-	 * column in each row.
+	 * column in each row.  We don't want to collect additional
+	 * footnotes, as we're going to do so in the next iteration, and
+	 * keep the current size (which will otherwise advance).
 	 */
 
+	assert(!p->footoff);
+	p->footoff = 1;
+	footsz = p->footsz;
+
 	TAILQ_FOREACH(top, &n->children, entries) {
 		assert(top->type == LOWDOWN_TABLE_HEADER ||
 			top->type == LOWDOWN_TABLE_BODY);
@@ -1002,7 +1027,7 @@ rndr_table(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 				p->last_blank = 0;
 				p->maxcol = SIZE_MAX;
 				p->col = 1;
-				if (!rndr(celltmp, mq, p, cell))
+				if (!rndr(celltmp, p, cell))
 					goto out;
 				if (widths[i] < p->col)
 					widths[i] = p->col;
@@ -1012,6 +1037,12 @@ rndr_table(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 			}
 	}
 
+	/* Restore footnotes. */
+
+	p->footsz = footsz;
+	assert(p->footoff);
+	p->footoff = 0;
+
 	/* Now actually print, row-by-row into the output. */
 
 	TAILQ_FOREACH(top, &n->children, entries) {
@@ -1029,7 +1060,7 @@ rndr_table(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 				p->last_blank = 0;
 				p->maxcol = SIZE_MAX;
 				p->col = 1;
-				if (!rndr(celltmp, mq, p, cell))
+				if (!rndr(celltmp, p, cell))
 					goto out;
 				assert(widths[i] >= p->col);
 				sz = widths[i] - p->col;
@@ -1138,16 +1169,15 @@ out:
 }
 
 static int
-rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
-	struct term *p, const struct lowdown_node *n)
+rndr(struct lowdown_buf *ob, struct term *p,
+	const struct lowdown_node *n)
 {
 	const struct lowdown_node	*child, *nn;
-	struct lowdown_meta		*m;
 	struct lowdown_buf		*metatmp;
+	void				*pp;
 	int32_t				 entity;
 	size_t				 i, col, vs;
 	ssize_t			 	 last_blank;
-	int				 rc;
 
 	/* Current nodes we're servicing. */
 
@@ -1175,8 +1205,6 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 	case LOWDOWN_BLOCKQUOTE:
 	case LOWDOWN_DEFINITION:
 	case LOWDOWN_DEFINITION_TITLE:
-	case LOWDOWN_FOOTNOTES_BLOCK:
-	case LOWDOWN_FOOTNOTE_DEF:
 	case LOWDOWN_HEADER:
 	case LOWDOWN_LIST:
 	case LOWDOWN_TABLE_BLOCK:
@@ -1218,12 +1246,6 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 	/* Output leading content. */
 
 	switch (n->type) {
-	case LOWDOWN_FOOTNOTES_BLOCK:
-		hbuf_truncate(p->tmp);
-		if (!hbuf_puts(p->tmp, ifx_foot) ||
-		    !rndr_buf(p, ob, n, p->tmp, &sty_foot))
-			return 0;
-		break;
 	case LOWDOWN_SUPERSCRIPT:
 		hbuf_truncate(p->tmp);
 		if (!hbuf_puts(p->tmp, ifx_super) ||
@@ -1238,44 +1260,6 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 		if (!hbuf_puts(p->tmp, ifx_meta_key) ||
 		    !rndr_buf(p, ob, n, p->tmp, &sty_meta_key))
 			return 0;
-		if (mq == NULL)
-			break;
-
-		/*
-		 * Manually render the children of the meta into a
-		 * buffer and use that as our value.  Start by zeroing
-		 * our terminal position and using another output buffer
-		 * (p->tmp would be clobbered by children).
-		 */
-
-		last_blank = p->last_blank;
-		p->last_blank = -1;
-		col = p->col;
-		p->col = 0;
-		m = calloc(1, sizeof(struct lowdown_meta));
-		if (m == NULL)
-			return 0;
-		TAILQ_INSERT_TAIL(mq, m, entries);
-		m->key = strndup(n->rndr_meta.key.data,
-			n->rndr_meta.key.size);
-		if (m->key == NULL)
-			return 0;
-		if ((metatmp = hbuf_new(128)) == NULL)
-			return 0;
-		TAILQ_FOREACH(child, &n->children, entries) {
-			p->stackpos++;
-			if (!rndr(metatmp, mq, p, child)) {
-				hbuf_free(metatmp);
-				return 0;
-			}
-			p->stackpos--;
-		}
-		m->value = strndup(metatmp->data, metatmp->size);
-		hbuf_free(metatmp);
-		if (m->value == NULL)
-			return 0;
-		p->last_blank = last_blank;
-		p->col = col;
 		break;
 	default:
 		break;
@@ -1283,38 +1267,72 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 
 	/* Descend into children. */
 
-	if (n->type != LOWDOWN_TABLE_BLOCK) {
+	switch (n->type) {
+	case LOWDOWN_FOOTNOTE:
+		if (p->footoff) {
+			p->footsz++;
+			break;
+		}
+		last_blank = p->last_blank;
+		p->last_blank = -1;
+		col = p->col;
+		p->col = 0;
+		if ((metatmp = hbuf_new(128)) == NULL)
+			return 0;
 		TAILQ_FOREACH(child, &n->children, entries) {
 			p->stackpos++;
-			if (!rndr(ob, mq, p, child))
+			if (!rndr(metatmp, p, child))
 				return 0;
 			p->stackpos--;
 		}
-	} else if (!rndr_table(ob, mq, p, n))
-		return 0;
+		p->last_blank = last_blank;
+		p->col = col;
+		pp = recallocarray(p->foots, p->footsz,
+			p->footsz + 1, sizeof(struct lowdown_buf *));
+		if (pp == NULL)
+			return 0;
+		p->foots = pp;
+		p->foots[p->footsz++] = metatmp;
+		break;
+	case LOWDOWN_TABLE_BLOCK:
+		if (!rndr_table(ob, p, n))
+			return 0;
+		break;
+	default:
+		TAILQ_FOREACH(child, &n->children, entries) {
+			p->stackpos++;
+			if (!rndr(ob, p, child))
+				return 0;
+			p->stackpos--;
+		}
+		break;
+	}
 
 	/* Output content. */
 
-	rc = 1;
 	switch (n->type) {
 	case LOWDOWN_HRULE:
 		hbuf_truncate(p->tmp);
 		if (!hbuf_puts(p->tmp, ifx_hrule))
 			return 0;
-		rc = rndr_buf(p, ob, n, p->tmp, NULL);
+		if (!rndr_buf(p, ob, n, p->tmp, NULL))
+			return 0;
 		break;
-	case LOWDOWN_FOOTNOTE_REF:
+	case LOWDOWN_FOOTNOTE:
 		hbuf_truncate(p->tmp);
 		if (!hbuf_printf(p->tmp, "%s%zu%s", ifx_fref_left,
-		    n->rndr_footnote_ref.num, ifx_fref_right))
+		    p->footsz, ifx_fref_right))
+			return 0;
+		if (!rndr_buf(p, ob, n, p->tmp, &sty_fref))
 			return 0;
-		rc = rndr_buf(p, ob, n, p->tmp, NULL);
 		break;
 	case LOWDOWN_RAW_HTML:
-		rc = rndr_buf(p, ob, n, &n->rndr_raw_html.text, NULL);
+		if (!rndr_buf(p, ob, n, &n->rndr_raw_html.text, NULL))
+			return 0;
 		break;
 	case LOWDOWN_MATH_BLOCK:
-		rc = rndr_buf(p, ob, n, &n->rndr_math.text, NULL);
+		if (!rndr_buf(p, ob, n, &n->rndr_math.text, NULL))
+			return 0;
 		break;
 	case LOWDOWN_ENTITY:
 		entity = entity_find_iso(&n->rndr_entity.text);
@@ -1322,19 +1340,25 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 			hbuf_truncate(p->tmp);
 			if (!rndr_entity(p->tmp, entity))
 				return 0;
-			rc = rndr_buf(p, ob, n, p->tmp, NULL);
-		} else
-			rc = rndr_buf(p, ob, n, &n->rndr_entity.text,
-				&sty_bad_ent);
+			if (!rndr_buf(p, ob, n, p->tmp, NULL))
+				return 0;
+		} else {
+			if (!rndr_buf(p, ob, n, 
+			     &n->rndr_entity.text, &sty_bad_ent))
+				return 0;
+		}
 		break;
 	case LOWDOWN_BLOCKCODE:
-		rc = rndr_buf(p, ob, n, &n->rndr_blockcode.text, NULL);
+		if (!rndr_buf(p, ob, n, &n->rndr_blockcode.text, NULL))
+			return 0;
 		break;
 	case LOWDOWN_BLOCKHTML:
-		rc = rndr_buf(p, ob, n, &n->rndr_blockhtml.text, NULL);
+		if (!rndr_buf(p, ob, n, &n->rndr_blockhtml.text, NULL))
+			return 0;
 		break;
 	case LOWDOWN_CODESPAN:
-		rc = rndr_buf(p, ob, n, &n->rndr_codespan.text, NULL);
+		if (!rndr_buf(p, ob, n, &n->rndr_codespan.text, NULL))
+			return 0;
 		break;
 	case LOWDOWN_LINK_AUTO:
 		if (p->opts & LOWDOWN_TERM_SHORTLINK) {
@@ -1342,9 +1366,13 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 			if (!hbuf_shortlink
 			    (p->tmp, &n->rndr_autolink.link))
 				return 0;
-			rc = rndr_buf(p, ob, n, p->tmp, NULL);
-		} else
-			rc = rndr_buf(p, ob, n, &n->rndr_autolink.link, NULL);
+			if (!rndr_buf(p, ob, n, p->tmp, NULL))
+				return 0;
+		} else {
+			if (!rndr_buf(p, ob, n,
+			     &n->rndr_autolink.link, NULL))
+				return 0;
+		}
 		break;
 	case LOWDOWN_LINK:
 		if (p->opts & LOWDOWN_TERM_NOLINK)
@@ -1359,9 +1387,13 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 			if (!hbuf_shortlink
 			    (p->tmp, &n->rndr_link.link))
 				return 0;
-			rc = rndr_buf(p, ob, n, p->tmp, NULL);
-		} else
-			rc = rndr_buf(p, ob, n, &n->rndr_link.link, NULL);
+			if (!rndr_buf(p, ob, n, p->tmp, NULL))
+				return 0;
+		} else {
+			if (!rndr_buf(p, ob, n,
+			     &n->rndr_link.link, NULL))
+				return 0;
+		}
 		break;
 	case LOWDOWN_IMAGE:
 		if (!rndr_buf(p, ob, n, &n->rndr_image.alt, NULL))
@@ -1375,16 +1407,20 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 		}
 		if (p->opts & LOWDOWN_TERM_NOLINK) {
 			hbuf_truncate(p->tmp);
-			if (!hbuf_puts(p->tmp, ifx_imgbox_left) ||
-			    !hbuf_puts(p->tmp, ifx_imgbox_right))
+			if (!hbuf_puts(p->tmp, ifx_imgbox_left))
+				return 0;
+			if (!hbuf_puts(p->tmp, ifx_imgbox_right))
+				return 0;
+			if (!rndr_buf(p, ob, n, p->tmp, &sty_imgbox))
 				return 0;
-			rc = rndr_buf(p, ob, n, p->tmp, &sty_imgbox);
 			break;
 		}
 		hbuf_truncate(p->tmp);
-		if (!hbuf_puts(p->tmp, ifx_imgbox_left) ||
-		    !hbuf_puts(p->tmp, ifx_imgbox_sep) ||
-		    !rndr_buf(p, ob, n, p->tmp, &sty_imgbox))
+		if (!hbuf_puts(p->tmp, ifx_imgbox_left))
+			return 0;
+		if (!hbuf_puts(p->tmp, ifx_imgbox_sep))
+			return 0;
+		if (!rndr_buf(p, ob, n, p->tmp, &sty_imgbox))
 			return 0;
 		if (p->opts & LOWDOWN_TERM_SHORTLINK) {
 			hbuf_truncate(p->tmp);
@@ -1400,20 +1436,40 @@ rndr(struct lowdown_buf *ob, struct lowdown_metaq *mq,
 		hbuf_truncate(p->tmp);
 		if (!hbuf_puts(p->tmp, ifx_imgbox_right))
 			return 0;
-		rc = rndr_buf(p, ob, n, p->tmp, &sty_imgbox);
+		if (!rndr_buf(p, ob, n, p->tmp, &sty_imgbox))
+			return 0;
 		break;
 	case LOWDOWN_NORMAL_TEXT:
-		rc = rndr_buf(p, ob, n, &n->rndr_normal_text.text, NULL);
+		if (!rndr_buf(p, ob, n,
+		     &n->rndr_normal_text.text, NULL))
+			return 0;
 		break;
 	default:
 		break;
 	}
-	if (!rc)
-		return 0;
 
 	/* Trailing block spaces. */
 
 	if (n->type == LOWDOWN_ROOT) {
+		if (p->footsz) {
+			if (!rndr_buf_vspace(p, ob, n, 2))
+				return 0;
+			hbuf_truncate(p->tmp);
+			if (!hbuf_puts(p->tmp, pfx_body.text))
+				return 0;
+			if (!hbuf_puts(p->tmp, ifx_foot))
+				return 0;
+			if (!rndr_buf_literal(p, ob, n, p->tmp, &sty_foot))
+				return 0;
+			if (!rndr_buf_vspace(p, ob, n, 2))
+				return 0;
+			for (i = 0; i < p->footsz; i++) {
+				if (!hbuf_putb(ob, p->foots[i]))
+					return 0;
+				if (!HBUF_PUTSL(ob, "\n"))
+					return 0;
+			}
+		}
 		if (!rndr_buf_vspace(p, ob, n, 1))
 			return 0;
 		while (ob->size && ob->data[ob->size - 1] == '\n')
@@ -1435,17 +1491,13 @@ int
 lowdown_term_rndr(struct lowdown_buf *ob,
 	void *arg, const struct lowdown_node *n)
 {
-	struct term		*p = arg;
-	struct lowdown_metaq	 metaq;
-	int			 rc;
+	struct term	*st = arg;
+	int		 rc;
 
-	TAILQ_INIT(&metaq);
+	st->stackpos = 0;
 
-	p->stackpos = 0;
-
-	rc = rndr(ob, &metaq, p, n);
-
-	lowdown_metaq_free(&metaq);
+	rc = rndr(ob, st, n);
+	rndr_free_footnotes(st);
 	return rc;
 }
 
